function xout = apply_tonemapping(xin, method, type)

if(strcmp(method, 'Colorimetric reproduction'))
    xin(xin < 0.0) = 0.0;
    xin(xin > 1.0) = 1.0;
else
    
    xin = xin - min(min(xin(:)));
    if(strcmp(type, 'vis'))
        lowhigh = stretchlim(rgb2gray(xin), [0.0 0.99]);
    else
        lowhigh = stretchlim(xin, [0.0 0.99]);
    end
    %lowhigh = stretchlim(xin, [0.0 0.95])
    
    xin = imadjust(xin, [lowhigh(1);lowhigh(2)], [], 1.0);
    %xin(xin < 0.0) = 0.0;
    %xin(xin > 1.0) = 1.0;
    
    %minval = min(min(xin(:)));
    %xin = (xin - minval)./max(max(xin(:)-minval));
    
end

switch method
    case 'Linear'
        xout = xin;
    case 'Local (Retinex)'
        xout = local_tm(xin, 3.0, 1.0);
    case 'Histogram equalization'
        switch type
            case 'vis'
                for i=1:3
                    xout(:,:,i) = histeq(xin(:,:,i));
                end
            case 'nir'
                
                xout = histeq(xin);
        end
    case 'Colorimetric reproduction'
        %Gamma correction, colorimetric reproduction
        gamma = zeros(size(xin));
        gamma(xin<=0.0031308) = 12.92*xin(xin<=0.0031308);
        gamma(xin>0.0031308) = 1.055*xin(xin>0.0031308).^(1.0/2.4)-0.055;
        
        gamma(gamma < 0.0) = 0.0;
        gamma(gamma > 1.0) = 1.0;
        xout = gamma;
    otherwise
        error('Specified tone mapping method does not exist.')
end