// ArrayBase.h

#pragma once

#include "Random.h"

#include <iostream>
#include <sstream>
#include <iomanip>
#include <vector>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

using namespace std;

template <class T> 
class CArray
{
 public:
  // constructors
  CArray();
  ~CArray();

  CArray(int nData);
  CArray(int* dim, int nDims);

  CArray(const CArray &A);
  //CArray(const vector<T> &v); // disabled 
  //CArray(const vector<vector<T> > &m);

  CArray& operator=(const CArray &A);
  //CArray& operator=(const vector<T> &v); // disabled
  //CArray& operator=(const vector<vector<T> > &m);
  CArray& operator=(const T t);

  // create 
  void Create(int nData);
  void Create(int* dim, int nDims);
  void Create(const CArray<int> &size);

  void Clear(void); // free memory 

  // accessors
  T& operator[](int index);
  T& operator()(const CArray<int> &sub);
  T& operator()(int* sub);

  const T operator[](int index) const;
  const T operator()(const CArray<int> &sub) const;
  const T operator()(int* sub) const;
        
  CArray<int> size(void) const;
  CArray<int> stride(void) const;
  int nData(void) const;
  bool isEmpty(void) const;
  bool isNotEmpty(void) const;

  string Dump(void) const;
  string Dump(int precision) const;

  // utility
  void zeros(void);
  void Random(int mod); 
  void randu(float s=1.0f);
  void randn(float s=1.0f);

  void Make1D(void);
  void Make2D(void);
  void Make3D(void);

  int  sub2ind(const CArray<int> &A) const;
  void ind2sub(CArray<int> &A, int index) const;

  // internal utility
  int  sub2ind(int* sub) const;
  void ind2sub(int* sub, int index) const;

  void deleteData(void); 
  void setDataNull(void);

  // member variables
  T*   m_data;
  int* m_dims;
  int* m_stride;

  int m_nDims;
  int m_nData;
};

typedef CArray<float>         CArrayf;
typedef CArray<int>           CArrayi;
typedef CArray<unsigned char> CArrayu;
typedef CArray<bool>          CArrayb;

//
// Implementation
//

// constructors

template <class T>
CArray<T>::CArray()
{
    //cerr << "CArray::CArray()" << endl;
  setDataNull();
}

template <class T>
CArray<T>::~CArray()
{
  // cerr << "CArray::~CArray()" << endl;	   	 
  deleteData();
  setDataNull();
}

template <class T>
CArray<T>::CArray(int nData)
{
  //cerr << "CArray::CArray(int nData)\n";
    setDataNull();
    Create(nData);
}

template <class T>
CArray<T>::CArray(int* dim, int nDims)
{
  //cerr << "CArray::CArray(int* dim, int nDims)\n";
    setDataNull();
    Create(dim, nDims);
}

template <class T>
CArray<T>::CArray(const CArray<T> &A)
{
  //cerr << "CArray::CArray(const CArray &A)" << endl;
    setDataNull();
    this->operator=(A);
}

/*template <class T>
CArray<T>::CArray(const vector<T> &v)
{
  setDataNull();
  this->operator=(v);
}

template <class T>
CArray<T>::CArray(const vector<vector<T> > &m)
{
  setDataNull();
  this->operator=(m);
}
*/

template <class T>
CArray<T>& CArray<T>::operator=(const CArray<T> &A)
{
  // cerr << "CArray::operator=(const CArray &A)" << endl;
  Create(A.m_dims, A.m_nDims);
  memcpy(m_data, A.m_data, m_nData * sizeof(T)); 
  return *this;
}

template <class T>
CArray<T>& CArray<T>::operator=(const T t)
{
  // cerr << "template <class T> CArray<T>& CArray<T>::operator=(const T t)\n";
  for (int i = 0; i < m_nData; i++)
    m_data[i] = t;
  return *this;
}

// create

template <class T>
void CArray<T>::Create(int nData)
{
  // Create 1-dimensional array
 
  if (nData == 0)
    {
      deleteData();
      setDataNull();
      return;      
    }

  if (m_nDims == 1)
    {
      // reuse dims/stride arrays
    }
  else
    {
      if (m_dims != NULL)
	delete[] m_dims;
      if (m_stride != NULL)
	delete[] m_stride;
      
      m_nDims  = 1;
      m_dims   = new int[1];
      m_stride = new int[1];
    }

  m_dims[0] = nData;
  m_stride[0] = 1;

  if (m_nData == nData)
    {
      // reuse data array
    }
  else
    {
      if (m_data != NULL)
	delete[] m_data;
      
      m_nData = nData;
      m_data  = new T[nData];
    }
 
  zeros(); // array is initialised to 0
}

template <class T>
void CArray<T>::Create(int* dims, int nDims)
{
  // cerr << "CArray<T>::Create()" << endl;

  // Create n-dimensional array

  // empty array input
  if (nDims == 0)
    {
      deleteData();
      setDataNull();
      return;
    }

  // create dims array
  if (m_nDims == nDims)
    {
      //cerr << "CArray::Create -- reuse dims/stride array\n";
    }
  else
    {
      //cerr << "CArray::Create -- allocate dims/stride array\n";
      if (m_dims != NULL)
	delete[] m_dims;
      if (m_stride != NULL)
	delete[] m_stride;

      m_nDims  = nDims;
      m_dims   = new int[nDims];
      m_stride = new int[nDims];
    }

  // copy dims array
  for (int i = 0; i < nDims; i++)
    m_dims[i] = dims[i];

  // create stride array
  int stride = 1;
  for (int i = nDims - 1; i >= 0; i--)
    {
      m_stride[i] = stride;
      stride *= m_dims[i];
    }

  // compute nData
  int nData = stride;

  /*int nData = m_dims[0];
    for (int i = 1; i < m_nDims; i++)
    nData *= m_dims[i];*/

  // create data array
  if (m_nData == nData)
    {
      //cerr << "CArray::Create -- reusing memory\n";
    }
  else
    {
      //cerr << "CArray::Create -- allocate memory\n";
		
      if (m_data != NULL)
	delete[] m_data;

      m_nData = nData;
      m_data = new T[m_nData];
    }

  zeros(); // array is initialised to zero
}

template <class T>
void CArray<T>::Create(const CArray<int> &size)
{
  Create(size.m_data, size.m_nData);
}

template <class T>
void CArray<T>::Clear(void)
{
  deleteData();
  setDataNull();
}


// accessors

template <class T>
T& CArray<T>::operator[](int index)
{
  return m_data[index];
}

template <class T>
T& CArray<T>::operator()(const CArray<int> &sub)
{	
  int index = sub2ind(sub);
  return m_data[index];	
}

template <class T>
T& CArray<T>::operator()(int* sub)
{	
  int index = sub2ind(sub);
  return m_data[index];	
}

template <class T>
const T CArray<T>::operator[](int index) const
{
  return m_data[index];
}

template <class T>
const T CArray<T>::operator()(const CArray<int> &sub) const
{	
  int index = sub2ind(sub);
  return m_data[index];	
}

template <class T>
const T CArray<T>::operator()(int* dim) const
{	
  int index = sub2ind(dim);
  return m_data[index];	
}

template <class T>
CArray<int> CArray<T>::size(void) const
{
    CArray<int> sz;
    
    if (this->isEmpty())
    {
        sz.Create(1);
        sz[0] = 0;
        return sz;
    }
    
    sz.Create(m_nDims);

    for (int i = 0; i < m_nDims; i++)
        sz[i] = m_dims[i];
    
    return sz;
}

template <class T>
CArray<int> CArray<T>::stride(void) const
{
  CArray<int> stride(m_nDims);
  
  for (int i = 0; i < m_nDims; i++)
    stride[i] = m_stride[i];

  return stride;
}

template <class T>
int CArray<T>::nData(void) const
{
    return m_nData;
}

template <class T>
bool CArray<T>::isEmpty(void) const
{
  return (m_nData == 0);
}

template <class T>
bool CArray<T>::isNotEmpty(void) const
{
    return (m_nData > 0);
}

// utility

template <class T>
void CArray<T>::zeros(void)
{
  memset(m_data, 0, m_nData * sizeof(T));
}

template <class T>
void CArray<T>::Random(int mod)
{
  for (int i = 0; i < m_nData; i++)
    {
      m_data[i] = (T)(rand()%mod);
    }
}

template <class T>
void CArray<T>::randu(float s)
{
    for (int i = 0; i < m_nData; i++)
    {
        m_data[i] = (T)(s * Randu());
    }
}

template <class T>
void CArray<T>::randn(float s)
{
    for (int i = 0; i < m_nData; i++)
    {
        m_data[i] = (T)(s * Randn());
    }
}

template <class T>
void CArray<T>::Make1D(void)
{
    if (m_nDims == 1)
        return;

    if (m_nDims > 1)
    {
        delete m_dims;
        m_dims = new int[1];
        delete m_stride;
        m_stride = new int[1];

        m_dims[0] = m_nData;
        m_stride[0] = 1;        
        m_nDims = 1;
    }
}

template <class T>
void CArray<T>::Make2D(void)
{
    if (m_nDims == 2)
        return;
    
    if (m_nDims > 2) // reduce to 2 dimensions
    {
        int nRows = m_dims[0];

        int nCols = 1; // if dim[3] etc >= 1 need more cols to compensate
        for (int i = 1; i < m_nDims; i++) 
            nCols *= m_dims[i]; 
        
        delete m_dims;
        m_dims = new int[2];
        delete m_stride;
        m_stride = new int[2];

        m_dims[0] = nRows;
        m_dims[1] = nCols;
        
        m_stride[0] = nCols;
        m_stride[1] = 1;

        m_nDims = 2;
    }

    if (m_nDims == 1) // expand to 2 dimensions
    {        
        delete m_dims;
        m_dims = new int[2];
        delete m_stride;
        m_stride = new int[2];
        
        m_dims[0] = 1;
        m_dims[1] = m_nData;
        
        m_stride[0] = m_nData;
        m_stride[1] = 1;

        m_nDims = 2;
    }
}

template <class T>
void CArray<T>::Make3D(void)
{
    if (m_nDims == 3) 
        return;

    if (m_nDims == 1)
    {
        delete m_dims;
        m_dims = new int[3];
        delete m_stride;
        m_stride = new int[3];
        
        m_dims[0] = 1;
        m_dims[1] = m_nData;
        m_dims[2] = 1;
        
        m_stride[0] = m_nData;
        m_stride[1] = 1;
        m_stride[2] = 1;

        m_nDims = 3;
    }

    if (m_nDims == 2)
    {
        int nRows = m_dims[0];
        int nCols = m_dims[1];
        
        delete m_dims;
        m_dims = new int[3];
        delete m_stride;
        m_stride = new int[3];
        
        m_dims[0] = nRows;
        m_dims[1] = nCols;
        m_dims[2] = 1;
        
        m_stride[0] = nCols;
        m_stride[1] = 1;
        m_stride[2] = 1;

        m_nDims = 3;
    }
        
    if (m_nDims > 3) // reduce to 3 dimensions
    {
        int nRows = m_dims[0];
        int nCols = m_dims[1];
        
        int nBands = 1;
        for (int i = 2; i < m_nDims; i++)
            nBands *= m_dims[i];
        
        delete m_dims;
        m_dims = new int[3];
        delete m_stride;
        m_stride = new int[3];

        m_dims[0] = nRows;
        m_dims[1] = nCols;
        m_dims[2] = nBands;
        
        m_stride[0] = nCols * nBands;
        m_stride[1] = nBands;
        m_stride[2] = 1;
        
        m_nDims = 3;
    }
        
}
    
template <class T>
int CArray<T>::sub2ind(const CArray<int> &A) const
{
  int offset = 0;

  for (int i = m_nDims - 1; i >= 0; i--)
    offset += A[i] * m_stride[i];

  return offset;
}

template <class T>
void CArray<T>::ind2sub(CArray<int> &A, int index) const
{
  int indexRemain = index;

  for (int i = 0; i < m_nDims; i++)
    {
      A[i] = indexRemain / m_stride[i];
      indexRemain -= A[i] * m_stride[i];
    }

}

// internal utility

template <class T>
int CArray<T>::sub2ind(int* sub) const
{
  // convert subscripts to array offset
  int offset = 0;

  for (int i = m_nDims - 1; i >= 0; i--)
    offset += sub[i] * m_stride[i];

  return offset;
}

template <class T>
void CArray<T>::ind2sub(int* sub, int index) const
{
  int indexRemain = index;

  for (int i = 0; i < m_nDims; i++)
    {
      sub[i] = indexRemain / m_stride[i];
      indexRemain -= sub[i] * m_stride[i];
    }

}

template <class T>
void CArray<T>::setDataNull(void)
{
  // Set all data to NULL
  // used for constructors
  m_data   = NULL;
  m_dims   = NULL;
  m_stride = NULL;
  m_nData  = 0;
  m_nDims  = 0;
}

template <class T>
void CArray<T>::deleteData(void)
{
  // Delete all data
  // use for constructor/descrutors
  if (m_data != NULL)
    delete[] m_data;
  if (m_dims != NULL)
    delete[] m_dims;
  if (m_stride != NULL)
    delete[] m_stride;
}  

template <class T>
string CArray<T>::Dump(void) const
{
    return this->Dump(3);
}

template <class T>
string CArray<T>::Dump(int precision) const
{
  stringstream os;
  
  //int precision = 3;
  // TODO: specialise this for ints etc.
  int width = 6;
  
  os << setprecision(precision);
  
  // TODO: these are not sticky
  //os << setfill(' ');
  //os << setw(7);

  const CArray<T> &A = *this;
  
  if (A.m_nDims == 0)
    return os.str();

  if (A.m_nDims == 1)
    {
      for (int i = 0; i < A.m_nData; i++)
	{
            //os << setw(width) << A.m_data[i] << " ";
            os << A.m_data[i] << " ";
	}
      os << endl;

      return os.str();
    }
  
 
  if (A.m_nDims == 2)
    {
      int i = 0;

      for (int r = 0; r < A.m_dims[0]; r++)
	{
	  for (int c = 0; c < A.m_dims[1]; c++)
	    {
                //os << setw(width) << A.m_data[i] << " ";
                os << A.m_data[i] << " ";
	      i++;
	    }
	  os << endl; 
	}
      return os.str();
    }

  if (A.m_nDims == 3)
    {
      int i = 0; 
      
      for (int r = 0; r < A.m_dims[0]; r++)
	{
	  for (int c = 0; c < A.m_dims[1]; c++)
	    {
	      os << "[";
	      for (int b = 0; b < A.m_dims[2]; b++)
              {
                  //os << setw(width) << A.m_data[i];
                  os  << A.m_data[i];
		  if (b < A.m_dims[2] - 1)
		    os << " ";
		  i++;
		}
	      os << "] ";
	    }
	  os << endl;
	}
      return os.str();
    }

  os << "CArray<T>::Dump() -- m_nDims = " << A.m_nDims << " not implemented\n";

  /*CCount C;
  C.Create(A.size());

  for (int i = 0; i < A.m_nData; i++)
    {
      cerr << "[";
      for (int j = 0; j < C.m_count.m_nData; j++)
	{
	  os << C.m_count.m_data[j];
	  if (j < C.m_count.m_nData - 1)
	    os << " ";
	}
      cerr << "] = ";
      cerr << A.m_data[i] << endl;
      C.IncCount();
      }*/
        
  return os.str();
}

