// ArrayFun.h

#pragma once

#include <iostream>
#include <math.h>
#include <vector>
#include "assert.h"
#include <algorithm>

#include "Count.h"
#include "ArrayCount.h"

using namespace std;

// unary array operations

template <class T> CArray<T> operator-(const CArray<T> &A1); // Unary minus

template <class T> void Not(CArray<T> &A, const CArray<T> &A1); // Logical not
template <class T> CArray<T> Not(const CArray<T> &A1); 
template <class T> CArray<T> operator!(const CArray<T> &A1); 

template <class T> void Abs(CArray<T> &A, const CArray<T> &A1); // Absolute value
template <class T> CArray<T> Abs(const CArray<T> &A); 
template <class T> void AbsInPlace(CArray<T> &A);

// binary array operations  (pass value, return value, in place)

template <class T> void Swap(CArray<T> &A1, CArray<T> &A2);  // Swap A1 and A2
template <class T> void Steal(CArray<T> &A1, CArray<T> &A2); // Swap A1 and A2 and clear A2

template <class T> void Add(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Add
template <class T> CArray<T> operator+(const CArray<T> &A1, const CArray<T> &A2); 
template <class T> void operator+=(CArray<T> &A, const CArray<T> &A1);

template <class T> void Subtract(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Subtract
template <class T> CArray<T> operator-(const CArray<T> &A1, const CArray<T> &A2); 
template <class T> void operator-=(CArray<T> &A, const CArray<T> &A1);

template <class T> void ScalarMult(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Scalar multiply
template <class T> CArray<T> ScalarMult(const CArray<T> &A1, const CArray<T> &A2);

template <class T> void ScalarDiv(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Scalar divide
template <class T> CArray<T> ScalarDiv(const CArray<T> &A1, const CArray<T> &A2);

template <class T> void Equal(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Elementwise equal
template <class T> CArray<T> operator==(const CArray<T> &A1, const CArray<T> &A2); 

template <class T> void And(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // And
template <class T> CArray<T> And(const CArray<T> &A1, const CArray<T> &A2); 

template <class T> void Or(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Or
template <class T> CArray<T> Or(const CArray<T> &A1, const CArray<T> &A2); 

template <class T> void XOr(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // XOr
template <class T> CArray<T> XOr(const CArray<T> &A1, const CArray<T> &A2); 

template <class T> bool isEqual(const CArray<T> &A1, const CArray<T> &A2); // All elements equal
template <class T> bool operator==(const CArray<T> &A1, const CArray<T> &A2);

template <class T> bool sizeEqual(const CArray<T> &A1, const CArray<T> &A2); // Sizes equal

template <class T> void Greater(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Greater than
template <class T> CArray<T> operator>(const CArray<T> &A1, const CArray<T> &A2); 

template <class T> void Less(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Less than
template <class T> CArray<T> operator<(const CArray<T> &A1, const CArray<T> &A2); 

template <class T> void Max(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Max
template <class T> CArray<T> Max(const CArray<T> &A1, const CArray<T> &A2); 

template <class T> void Min(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Min
template <class T> CArray<T> Min(const CArray<T> &A1, const CArray<T> &A2); 

template <class T> void Concatenate(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2); // Concatenate
template <class T> CArray<T> Concatenate(const CArray<T> &A1, const CArray<T> &A2);

// array scalar operations  (pass value, return value, in place)

template <class T> void Add(CArray<T> &A, const CArray<T> &A1, const T t); // Add
template <class T> CArray<T> operator+(const CArray<T> &A, const T t);
template <class T> void operator+=(CArray<T> &A, const T t); 

template <class T> void Subtract(CArray<T> &A, const CArray<T> &A1, const T t); // Subtract
template <class T> CArray<T> operator-(const CArray<T> &A, const T t);
template <class T> void operator-=(CArray<T> &A, const T t); 

template <class T> void Multiply(CArray<T> &A, const CArray<T> &A1, const T t); // Multiply
template <class T> CArray<T> operator*(const CArray<T> &A, const T t);
template <class T> void operator*=(CArray<T> &A, const T t); 

template <class T> void Divide(CArray<T> &A, const CArray<T> &A1, const T t); // Divide
template <class T> CArray<T> operator/(const CArray<T> &A, const T t);
template <class T> void operator/=(CArray<T> &A, const T t); 

template <class T> void Greater(CArray<T> &A, const CArray<T> &A1, const T t); // Greater than
template <class T> CArray<T> operator>(const CArray<T> &A1, const T t); 

template <class T> void Less(CArray<T> &A, const CArray<T> &A1, const T t); // Less than
template <class T> CArray<T> operator<(const CArray<T> &A1, const T t); 

template <class T> void Max(CArray<T> &A, const CArray<T> &A1, const T t); // Max
template <class T> CArray<T> Max(const CArray<T> &A1, const T t); 

template <class T> void Min(CArray<T> &A, const CArray<T> &A1, const T t); // Min
template <class T> CArray<T> Min(const CArray<T> &A1, const T t); 

template <class T> void Equal(CArray<T> &A, const CArray<T> &A1, const T t); // Equal to
template <class T> CArray<T> operator==(const CArray<T> &A1, const T t); 

template <class T> void Pow(CArray<T> &A, const CArray<T> &A1, const T t); // Scalar power
template <class T> CArray<T> Pow(const CArray<T> &A1, const T t);
template <class T> CArray<T> operator^(const CArray<T> &A1, const T t);

template <class T> void Pow2(CArray<T> &A, const CArray<T> &A1); // Scalar square
template <class T> CArray<T> Pow2(const CArray<T> &A1);

template <class T> void Sqrt(CArray<T> &A, const CArray<T> &A1); // Scalar square
template <class T> CArray<T> Sqrt(const CArray<T> &A1);

template <class T> void Log(CArray<T> &A, const CArray<T> &A1); // Scalar square
template <class T> CArray<T> Log(const CArray<T> &A1);

// convolve, copy, select, find

template <class T> void Convolve(T* dstptr, T* srcptr, int nData, const CArray<T> &kernel); // Convolution

template <class T> void Convolve1D(CArray<T> &A, const CArray<T> &A1, const CArray<T> &kernel); // 1D Convolution
template <class T> CArray<T> Convolve1D(const CArray<T> &A1, const CArray<T> &kernel);

template <class T> void Copy(CArray<T> &Adst, const CArrayi &dstStart, 
                                  const CArray<T> &Asrc, const CArrayi &srcStart, const CArrayi &srcEnd); // Array copy
template <class T> void Copy(CArray<T> &Adst, int dstStart, const CArray<T> &Asrc, int srcStart, int srcEnd);
template <class T> void Copy(CArray<T> &Adst, int dstStart, const CArray<T> &Asrc);

template <class T> void Add(CArray<T> &Adst, int dstStart, const CArray<T> &Asrc, int srcStart, int srcEnd);
template <class T> void Add(CArray<T> &Adst, int dstStart, const CArray<T> &Asrc);

template <class T> void Select(CArray<T> &Adst, const CArray<T> &Asrc, const CArrayi &srcStart, const CArrayi &srcEnd);
template <class T> CArray<T> Select(const CArray<T> &Asrc, const CArrayi &srcStart, const CArrayi &srcEnd);
template <class T> void Select(CArray<T> &Adst, const CArray<T> &Asrc, int idx0, int idx1);
template <class T> CArray<T> Select(const CArray<T> &Asrc, int idx0, int idx1);

template <class T> void SelectIndex(CArray<T> &Adst, const CArray<T> &Asrc, const CArrayi &index);
template <class T> CArray<T> SelectIndex(const CArray<T> &Asrc, const CArrayi &index);

// find, sort

template <class T> void Find(vector<int> &index, vector<T> &value, const CArray<T> &A);
template <class T> void Findv(vector<T> &value, const CArray<T> &A);
template <class T> void Findi(vector<int> &index, const CArray<T> &A);

template <class T> void Find(CArrayi &index, CArray<T> &value, const CArray<T> &A);
template <class T> void Findi(CArrayi &index, const CArray<T> &A);
template <class T> void Findv(CArray<T> &value, const CArray<T> &A);
template <class T> CArrayi Findi(const CArray<T> &A);
template <class T> CArray<T> Findv(const CArray<T> &A);

template <class T> void Index(CArray<T> &A, const CArrayi &indicies, int N);
template <class T> void Union(CArray<T> &A, const CArrayi &set1, const CArrayi &set2);
template <class T> void Intersect(CArray<T> &A, const CArrayi &set1, const CArrayi &set2);

template <class T> int FirstNonZero(const CArray<T> &A);
template <class T> int LastNonZero(const CArray<T> &A);

template <class T> class CSort {public: T value; int index;};
//template <class T> bool operator>(CSort<T> s1, CSort<T> s2) { return (s1.value > s2.value); }
//template <class T> bool operator<(CSort<T> s1, CSort<T> s2) { return (s1.value < s2.value); }
template <class T> void Swap(CSort<T> &s1, CSort<T> &s2);
template <class T> int QSortPartition(vector<CSort<T> > &vs, int left, int right, int pivotIdx, bool ascending = true);
template <class T> void QSort(vector<CSort<T> > &vs, int left, int right, bool ascending = true);

template <class T> void Sort(CArrayi &sortIndex, CArray<T> &sortValue, const CArray<T> &arr, bool ascending = true);
template <class T> void SortIndex(CArrayi &sortIndex, const CArray<T> &arr, bool ascending = true);
template <class T> CArrayi SortIndex(const CArray<T> &arr, bool ascending = true);
template <class T> void SortValue(CArray<T> &sortValue, const CArray<T> &arr, bool ascending = true);
template <class T> CArray<T> SortValue(const CArray<T> &arr, bool ascending = true);

// array manipulation

template <class T> T Sum(const CArray<T> &A);    // sum
template <class T> T Sumsq(const CArray<T> &A);  // sum squares
template <class T> T Mean(const CArray<T> &A);   // mean
template <class T> T Median(const CArray<T> &A); // median
template <class T> T Var(const CArray<T> &A);    // variance
template <class T> T StdDev(const CArray<T> &A); // standard deviation
template <class T> T Norm(const CArray<T> &A);   // norm

template <class T> void Max(int &maxIdx, T &maxVal, const CArray<T> &A); // max index and value
template <class T> void Max(CArrayi &maxSubs, T &maxVal, const CArray<T> &A); // max subscripts and value
template <class T> int Maxi(const CArray<T> &A);
template <class T> T Max(const CArray<T> &A);    

template <class T> T Min(const CArray<T> &A);    // min

template <class T> void NormaliseRange(CArray<T> &Adst, const CArray<T> &Asrc, T min, T max); // linearly map array to range min/max
template <class T> CArray<T> NormaliseRange(const CArray<T> &Asrc, T mn, T mx);
template <class T> void NormaliseCentred(CArray<T> &Adst, const CArray<T> &Asrc, T min, T max); // normalise mapping 0 to (min + max) / 2
template <class T> CArray<T> NormaliseCentred(const CArray<T> &Asrc, T mn, T mx);
template <class T> void NormaliseUnit(CArray<T> &Adst, const CArray<T> &Asrc); // normalise to unit length
template <class T> CArray<T> NormaliseUnit(const CArray<T> &Asrc);

template <class T> void ClipMinMax(CArray<T> &A, const T mn, const T mx);

// convert to std vector

template <class T> void Convert(vector<T> &v, const CArray<T> &A);
template <class T> void Convert(CArray<T> &A, const vector<T> &v);

// input/output

template <class T> ostream& operator<<(ostream &os, const CArray<T> &A); // machine readable output
template <class T> istream& operator>>(istream &is, CArray<T> &A);       // machine readable input

template <class T> void Write(const CArray<T> &arr, string filename); // binary file output
template <class T> void Read(CArray<T> &arr, string filename);        // binary file input

//
// Implementation
//

// unary array operations

template <class T>
CArray<T> operator-(const CArray<T> &A1)
{
  CArray<T> M = A1 * -1.0f;
  return(M);
}

template <class T>
void Not(CArray<T> &A, const CArray<T> &A1)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A1.m_nData; i++)
    A.m_data[i] = !(A1.m_data[i]);
}

template <class T>
CArray<T> Not(const CArray<T> &A1)
{
  CArray<T> A;
  Not(A, A1);
  return(A);
}

template <class T>
CArray<T> operator!(const CArray<T> &A1)
{
  CArray<T> A;
  Not(A, A1);
  return(A);
}

template <class T>
void Abs(CArray<T> &A, const CArray<T> &A1)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A1.m_nData; i++)
    A.m_data[i] = (T)fabs(A1.m_data[i]); // Note: overide with abs for int/byte
}


template <class T>
CArray<T> Abs(const CArray<T> &A1)
{
  CArray<T> A;
  Abs(A, A1);
  return(A);
}

template <class T> 
void AbsInPlace(CArray<T> &A)
{
    for (int i = 0; i < A.m_nData; i++)
        A.m_data[i] = (T)fabs(A.m_data[i]);
}

// binary array operations

// Swap
template <class T> void Swap(CArray<T> &A1, CArray<T> &A2)
{
    // Swap arrays without copying data
    T*   A1_m_data   = A1.m_data;
    int* A1_m_dims   = A1.m_dims;
    int* A1_m_stride = A1.m_stride;
    int  A1_m_nDims  = A1.m_nDims;
    int  A1_m_nData  = A1.m_nData;

    A1.m_data   = A2.m_data;
    A1.m_dims   = A2.m_dims;
    A1.m_stride = A2.m_stride;
    A1.m_nDims  = A2.m_nDims;
    A1.m_nData  = A2.m_nData;

    A2.m_data   = A1_m_data;
    A2.m_dims   = A1_m_dims;
    A2.m_stride = A1_m_stride;
    A2.m_nDims  = A1_m_nDims;
    A2.m_nData  = A1_m_nData;
}

// Steal
template <class T> void Steal(CArray<T> &A1, CArray<T> &A2)
{
    // Swap A1 and A2 and clear A2
    Swap(A1, A2);
    A2.Clear();    
}

// Add
template <class T>
void Add(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] + A2.m_data[i];
}

template <class T>
CArray<T> operator+(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  Add(A, A1, A2);
  return A;
}

template <class T>
void operator+=(CArray<T> &A, const CArray<T> &A1)
{
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] += A1.m_data[i];
}

// Subtract
template <class T>
void Subtract(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] - A2.m_data[i];
}

template <class T>
CArray<T> operator-(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  Subtract(A, A1, A2);
  return A;
}

template <class T>
void operator-=(CArray<T> &A, const CArray<T> &A1)
{
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] -= A1.m_data[i];
}

// Scalar multiply
template <class T>
void ScalarMult(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] * A2.m_data[i];
}

template <class T>
CArray<T> ScalarMult(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  ScalarMult(A, A1, A2);
  return A;
}

// Scalar divide
template <class T>
void ScalarDiv(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] / A2.m_data[i];
}

template <class T>
CArray<T> ScalarDiv(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  ScalarDiv(A, A1, A2);
  return A;
}

// Equal
template <class T>
void Equal(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = (A1.m_data[i] == A2.m_data[i]);
}

template <class T>
CArray<T> operator==(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  Equal(A, A1, A2);
  return A;
}

// And
template <class T>
void And(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] && A2.m_data[i];
}

template <class T>
CArray<T> And(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  And(A, A1, A2);
  return A;
}

// Or
template <class T>
void Or(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] || A2.m_data[i];
}

template <class T>
CArray<T> Or(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  Or(A, A1, A2);
  return A;
}

// XOr
template <class T>
void XOr(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = !(A1.m_data[i] == A2.m_data[i]);
}

template <class T>
CArray<T> XOr(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  XOr(A, A1, A2);
  return A;
}

// isEqual
template <class T> 
bool isEqual(const CArray<T> &A1, const CArray<T> &A2)
{
    if (!sizeEqual(A1, A2))
        return false;
    
    for (int i = 0; i < A1.m_nData; i++)
    {
        if (A1.m_data[i] != A2.m_data[i])
            return false;
    }

    return true;
}

template <class T> 
bool operator==(const CArray<T> &A1, const CArray<T> &A2)
{
    return (isEqual(A1, A2));
}

// Size equal
template <class T> 
bool sizeEqual(const CArray<T> &A1, const CArray<T> &A2)
{
  if (A1.m_nDims != A2.m_nDims)
    return false;
  else
    {
      for (int i = 0; i < A1.m_nDims; i++)
	if (A1.m_dims[i] != A2.m_dims[i])
	  return false;
    }

  return true;
}

// Greater than
template <class T>
void Greater(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = (A1.m_data[i] > A2.m_data[i]);
}

template <class T>
CArray<T> operator>(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  Greater(A, A1, A2);
  return A;
}

// Less than
template <class T>
void Less(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = (A1.m_data[i] < A2.m_data[i]);
}

template <class T>
CArray<T> operator<(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  Less(A, A1, A2);
  return A;
}

// Max
template <class T>
void Max(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = max(A1.m_data[i],A2.m_data[i]);
}

template <class T>
CArray<T> Max(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  Max(A, A1, A2);
  return A;
}

// Min
template <class T>
void Min(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = min(A1.m_data[i],A2.m_data[i]);
}

template <class T>
CArray<T> Min(const CArray<T> &A1, const CArray<T> &A2)
{
  CArray<T> A;
  Min(A, A1, A2);
  return A;
}

template <class T> 
void Concatenate(CArray<T> &A, const CArray<T> &A1, const CArray<T> &A2)
{
    A.Create(A1.nData() + A2.nData());
    Copy(A, 0, A1);
    Copy(A, A1.nData(), A2);
}

template <class T> CArray<T> Concatenate(const CArray<T> &A1, const CArray<T> &A2)
{
    CArray<T> A;
    Concatenate(A, A1, A2);
    return A;
}

// array scalar operations

// Add
template <class T>
void Add(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] + t;
}

template <class T>
CArray<T> operator+(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Add(A, A1, t);
  return A;
}

template <class T>
void operator+=(CArray<T> &A, const T t)
{
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] += t;
}

// Subtract
template <class T>
void Subtract(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] - t;
}

template <class T>
CArray<T> operator-(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Subtract(A, A1, t);
  return A;
}

template <class T>
void operator-=(CArray<T> &A, const T t)
{
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] -= t;
}

// Multiply
template <class T>
void Multiply(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] * t;
}

template <class T>
CArray<T> operator*(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Multiply(A, A1, t);
  return A;
}

template <class T>
void operator*=(CArray<T> &A, const T t)
{
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] *= t;
}


// Divide
template <class T>
void Divide(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = A1.m_data[i] / t;
}

template <class T>
CArray<T> operator/(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Divide(A, A1, t);
  return A;
}

template <class T>
void operator/=(CArray<T> &A, const T t)
{
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] /= t;
}

// Greater
template <class T>
void Greater(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = (A1.m_data[i] > t);
}

template <class T>
CArray<T> operator>(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Greater(A, A1, t);
  return A;
}

// Less
template <class T>
void Less(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = (A1.m_data[i] < t);
}

template <class T>
CArray<T> operator<(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Less(A, A1, t);
  return A;
}

// Max
template <class T>
void Max(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = max(A1.m_data[i],t);
}

template <class T>
CArray<T> Max(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Max(A, A1, t);
  return A;
}

// Min
template <class T>
void Min(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = min(A1.m_data[i],t);
}

template <class T>
CArray<T> Min(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Min(A, A1, t);
  return A;
}

// Equal
template <class T>
void Equal(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = (A1.m_data[i] == t);
}

template <class T>
CArray<T> operator==(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Equal(A, A1, t);
  return A;
}

// Scalar Pow
template <class T>
void Pow(CArray<T> &A, const CArray<T> &A1, const T t)
{
  A.Create(A1.m_dims, A1.m_nDims);
  
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = (T)pow(A1.m_data[i], t);
}

template <class T>
CArray<T> operator^(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Pow(A, A1, t);
  return A;
}

template <class T>
CArray<T> Pow(const CArray<T> &A1, const T t)
{
  CArray<T> A;
  Pow(A, A1, t);
  return A;
}

// Scalar Pow2
template <class T>
void Pow2(CArray<T> &A, const CArray<T> &A1)
{
    ScalarMult(A, A1, A1);
}

template <class T>
CArray<T> Pow2(const CArray<T> &A1)
{
    CArray<T> A; 
    Pow2(A, A1);
    return A;
}

// Scalar Sqrt
template <class T>
void Sqrt(CArray<T> &A, const CArray<T> &A1)
{
    Pow(A, A1, 0.5f);
}

template <class T>
CArray<T> Sqrt(const CArray<T> &A1)
{
    CArray<T> A; 
    Sqrt(A, A1);
    return A;
}

// Scalar Log
template <class T>
void Log(CArray<T> &A, const CArray<T> &A1)
{
    A.Create(A1.m_dims, A1.m_nDims);

    for (int i = 0; i < A.m_nData; i++)
        A[i] = log(A1[i]);
}

template <class T>
CArray<T> Log(const CArray<T> &A1)
{
    CArray<T> A; 
    Log(A, A1);
    return A;
}

// convolve, copy and select

// Convolution
template <class T> 
void Convolve(T* dstptr, T* srcptr, int nData, const CArray<T> &kernel)
{    
    int ksize = kernel.m_nData;
    T* kptr0  = kernel.m_data;
    
    // make sure kernel length is odd
    CArray<T> kernel2;
    if (ksize % 2 == 0)
      {
	kernel2.Create(ksize + 1);
	for (int i = 0; i < kernel.m_nData; i++)
	  kernel2[i] = kernel[i];
	kptr0 = kernel2.m_data;
	ksize++;
      }
    
    int length = (ksize - 1)/2;

    // extend data outside range
    CArray<T> buffer(nData + 2 * length);
  
    // extend left
    T srcFirst = *srcptr;
    for (int i = 0; i < length; i++)
      buffer[i] = srcFirst;
	
    // copy centre
    memcpy(&buffer[length], srcptr, nData * sizeof(T));

    // extend right
    T srcLast = *(srcptr + nData - 1);
    for (int i = nData + length; i < nData + 2 * length; i++)
      buffer[i] = srcLast;
     
    // Convolve buffer with kernel    
    T *kptr;
    T *bufferptr;
    T *kptr1      = kptr0 + ksize - 1;
    T *bufferptr0 = buffer.m_data;

    for (int i = 0; i < nData; i++)
    {
        kptr      = kptr1;
        bufferptr = bufferptr0;

        for (int k = 0; k < ksize; k++)
        {
            *dstptr += *bufferptr * *kptr;
            kptr--;
            bufferptr++;
        }
        
        dstptr++;
        bufferptr0++;
    }

}

// 1D Convolution
template <class T> 
void Convolve1D(CArray<T> &A, const CArray<T> &A1, const CArray<T> &kernel)
{
    int nData = A1.m_nData;
    A.Create(A1.m_dims, A1.m_nDims);

    Convolve(A.m_data, A1.m_data, nData, kernel);
}

template <class T> CArray<T> Convolve1D(const CArray<T> &A1, const CArray<T> &kernel)
{
    CArray<T> A;
    Convolve1D(A, A1, kernel);
    return A;
}

// Copy
template <class T> 
void Copy(CArray<T> &Adst, const CArrayi &dstStart, 
                                  const CArray<T> &Asrc, const CArrayi &srcStart, const CArrayi &srcEnd0)
{
  
  CArrayi srcRange = srcEnd0 - srcStart + 1;
  CArrayi dstRange = Adst.size() - dstStart;

  CArrayi range; 
  range = Min(srcRange, dstRange);
  
  CArrayi dstEnd = dstStart + range - 1;
  CArrayi srcEnd = srcStart + range - 1;
  
  int nCount1;
  int nCount2;

  CArrayCount dstCount;
  nCount1 = dstCount.Create(dstStart, dstEnd, Adst.stride());

  CArrayCount srcCount;
  nCount2 = srcCount.Create(srcStart, srcEnd, Asrc.stride());
  
  assert(nCount1 == nCount2);

  for (int i = 0; i < nCount1; i++)
    {
      Adst[dstCount.m_idx] = Asrc[srcCount.m_idx];
      dstCount.IncCount();
      srcCount.IncCount();      
    }
  
}

template <class T> 
void Copy(CArray<T> &Adst, int dstStart, const CArray<T> &Asrc, int srcStart, int srcEnd)
{
    int nDataCopy = min(srcEnd - srcStart + 1, Adst.m_nData - dstStart);
    memcpy(&Adst.m_data[dstStart], &Asrc.m_data[srcStart], nDataCopy * sizeof(T));
}

template <class T> void Copy(CArray<T> &Adst, int dstStart, const CArray<T> &Asrc)
{
    Copy(Adst, dstStart, Asrc, 0, Asrc.m_nData - 1);
}

template <class T> 
void Add(CArray<T> &Adst, int dstStart, const CArray<T> &Asrc, int srcStart, int srcEnd)
{
    int nDataAdd = min(srcEnd - srcStart + 1, Adst.m_nData - dstStart);

    T* dstPtr = &Adst.m_data[dstStart];
    T* srcPtr = &Asrc.m_data[srcStart];
    
    for (int i = 0; i < nDataAdd; i++)
    {
        *dstPtr += *srcPtr;
        dstPtr++;
        srcPtr++;
    }
}

template <class T> void Add(CArray<T> &Adst, int dstStart, const CArray<T> &Asrc)
{
    Add(Adst, dstStart, Asrc, 0, Asrc.m_nData - 1);
}

// Select
template <class T> 
void Select(CArray<T> &Adst, const CArray<T> &Asrc, const CArrayi &srcStart, const CArrayi &srcEnd)
{
  CArrayi dstStart(srcStart.m_nData);
  CArrayi dstEnd = srcEnd - srcStart;
  Adst.Create(dstEnd + 1);
  
  int nCount1;
  int nCount2;

  CArrayCount dstCount;
  nCount1 = dstCount.Create(dstStart, dstEnd, Adst.stride());

  CArrayCount srcCount;
  nCount2 = srcCount.Create(srcStart, srcEnd, Asrc.stride());
  
  assert(nCount1 == nCount2);

  for (int i = 0; i < nCount1; i++)
    {
      Adst[dstCount.m_idx] = Asrc[srcCount.m_idx];
      dstCount.IncCount();
      srcCount.IncCount();      
    }  
}


template <class T> 
CArray<T> Select(const CArray<T> &Asrc, const CArrayi &srcStart, const CArrayi &srcEnd)
{
  CArray<T> Adst;
  Select(Adst, Asrc, srcStart, srcEnd);
  return Adst;
}

template <class T> void Select(CArray<T> &Adst, const CArray<T> &Asrc, int idx0, int idx1)
{
    int sz = idx1 - idx0 + 1;
    Adst.Create(sz);

    T* srcptr = Asrc.m_data + idx0;
    T* dstptr = Adst.m_data;

    memcpy(dstptr, srcptr, sz * sizeof(T));
}

template <class T> CArray<T> Select(const CArray<T> &Asrc, int idx0, int idx1)
{
    CArray<T> Adst;
    Select(Adst, Asrc, idx0, idx1);
    return Adst;
}

template <class T> void SelectIndex(CArray<T> &Adst, const CArray<T> &Asrc, const CArrayi &index)
{
    Adst.Create(index.m_nData);
    
    for (int i = 0; i < index.m_nData; i++)
        Adst[i] = Asrc[index.m_data[i]];
}

template <class T> CArray<T> SelectIndex(const CArray<T> &Asrc, const CArrayi &index)
{
    CArray<T> Adst;
    SelectIndex(Adst, Asrc, index);
    return Adst;
}

template <class T>
void Find(vector<int> &index, vector<T> &value, const CArray<T> &A)
{
  for (int i = 0; i < A.m_nData; i++)
    {
      float val = A[i];
      if (val != 0)
	{
	  index.push_back(i);
	  value.push_back(val);
	}
    }
}

template <class T> 
void Findv(vector<T> &value, const CArray<T> &A)
{
  for (int i = 0; i < A.m_nData; i++)
    {
      float val = A[i];
      if (val != 0)
	value.push_back(val);
    }
}

template <class T> 
void Findi(vector<int> &index, const CArray<T> &A)
{
  for (int i = 0; i < A.m_nData; i++)
    if (A[i] != 0)
      index.push_back(i);
}

template <class T>
void Find(CArrayi &index, CArray<T> &value, const CArray<T> &A)
{
    int n = 0;
    for (int i = 0; i < A.nData(); i++)
        if (A[i] != 0) n++;

    index.Create(n);
    value.Create(n);
    
    int j = 0;
    for (int i = 0; i < A.nData(); i++)
    {
        if (A[i] != 0)
        {
            index[j] = i;
            value[j] = A[i];
            j++;
        }
    }
}

template <class T>
void Findi(CArrayi &index, const CArray<T> &A)
{
    CArray<T> value;
    Find(index, value, A);
}

template <class T>
void Findv(CArray<T> &value, const CArray<T> &A)
{
    CArrayi index;
    Find(index, value, A);
}

template <class T>
CArrayi Findi(const CArray<T> &A)
{
    CArrayi index;
    CArray<T> value;
    Find(index, value, A);
    return index;
}

template <class T>
CArray<T> Findv(const CArray<T> &A)
{
    CArrayi index;
    CArray<T> value;
    Find(index, value, A);
    return value;
}

template <class T> 
void Index(CArray<T> &A, const CArrayi &indicies, int N)
{
    A.Create(N);
    for (int i = 0; i < indicies.nData(); i++)
        A[indicies[i]] = (T)1;
}

template <class T> void Union(CArray<T> &A, const CArrayi &set1, const CArrayi &set2)
{
    CArrayi index, index1, index2;
    int n = max(Max(set1), Max(set2)) + 1;
    Index(index1, set1, n);
    Index(index2, set2, n);
    Or(index, index1, index2);
    Findi(A, index);
}

template <class T> void Intersect(CArray<T> &A, const CArrayi &set1, const CArrayi &set2)
{
    CArrayi index, index1, index2;
    int n = max(Max(set1), Max(set2)) + 1;
    Index(index1, set1, n);
    Index(index2, set2, n);
    And(index, index1, index2);
    Findi(A, index);
}    
    
template <class T> 
int FirstNonZero(const CArray<T> &A)
{
  for (int i = 0; i < A.nData(); i++)
    {
      if (A[i] != 0)
	return i;
    }
  return -1;
}

template <class T>
int LastNonZero(const CArray<T> &A)
{
    for (int i = A.nData() - 1; i >= 0; i--)
    {
        if (A[i] != 0)
            return i;
    }
    return -1;
}

template <class T> void Swap(CSort<T> &s1, CSort<T> &s2)
{
    T   s1value = s1.value;
    int s1index = s1.index;
    
    s1.value = s2.value;
    s1.index = s2.index;
    
    s2.value = s1value;
    s2.index = s1index;
}
    
template <class T> int QSortPartition(vector<CSort<T> > &vs, int left, int right, int pivotIdx, bool ascending)
{
    T pivotVal = vs[pivotIdx].value;
    Swap(vs[right], vs[pivotIdx]);

    int swapIdx = left;
    for (int i = left; i < right; i++)
    {
        bool doSwap = false;       
        if (((vs[i].value < pivotVal) && ascending) || ((vs[i].value > pivotVal) && !ascending))
            doSwap = true;

        if (doSwap)
        {
            Swap(vs[i], vs[swapIdx]);
            swapIdx++;
        }  
    }
    Swap(vs[right], vs[swapIdx]);
    
    return swapIdx;
}

template <class T> void QSort(vector<CSort<T> > &vs, int left, int right, bool ascending)
{
    if (right > left)
    {
        int pivotIdx = (int)(right + left)/2;
        int pivotIdx2 = QSortPartition(vs, left, right, pivotIdx, ascending);
        QSort(vs, left, pivotIdx2 - 1, ascending);
        QSort(vs, pivotIdx2 + 1, right, ascending);    
    }
}

template <class T> 
void Sort(CArrayi &sortIndex, CArray<T> &sortValue, const CArray<T> &arr, bool ascending)
{
    vector<CSort<T> > vs;
    vs.resize(arr.nData());
    for (int i = 0; i < arr.nData(); i++)
    {
        vs[i].value = arr[i];
        vs[i].index = i;
    }
  
    QSort(vs, 0, (int)vs.size() - 1, ascending);

// The following does not compile in VStudio5
//    if (ascending)
//        sort(vs.begin(), vs.end(), less<CSort<T> >());        
//    else
//        sort(vs.begin(), vs.end(), greater<CSort<T> >());
    
    sortIndex.Create(arr.nData());
    sortValue.Create(arr.nData());

    for (int i = 0; i < arr.nData(); i++)
    {
        sortIndex[i] = vs[i].index;
        sortValue[i] = vs[i].value;        
    }
}    

template <class T>
void SortIndex(CArrayi &sortIndex, const CArray<T> &arr, bool ascending)
{
    CArray<T> sortValue;
    Sort(sortIndex, sortValue, arr, ascending);
}

template <class T>
CArrayi SortIndex(const CArray<T> &arr, bool ascending)
{
    CArrayi sortIndex;
    CArray<T> sortValue;
    Sort(sortIndex, sortValue, arr, ascending);
    return sortIndex;
}

template <class T>
void SortValue(CArray<T> &sortValue, const CArray<T> &arr, bool ascending)
{
    CArrayi sortIndex;
    Sort(sortIndex, sortValue, arr, ascending);
}

template <class T>
CArray<T> SortValue(const CArray<T> &arr, bool ascending)
{
  CArrayi sortIndex;
  CArray<T> sortValue;
  Sort(sortIndex, sortValue, arr, ascending);
  return sortValue;
}

// array manipulation

// Sum
template <class T>
T Sum(const CArray<T> &A)
{
  T s = 0;

  for (int i = 0; i < A.m_nData; i++)
    s += A.m_data[i];

  return s;
}

// Sumsq
template <class T>
T Sumsq(const CArray<T> &A)
{
  T sumsq = 0;

  for (int i = 0; i < A.m_nData; i++)
    sumsq += A.m_data[i] * A.m_data[i];

  return sumsq;
}

// Mean
template <class T>
T Mean(const CArray<T> &A)
{
    T s = Sum(A);
    return (s / (float)A.m_nData);
}

// Median
template <class T>
T Median(const CArray<T> &A)
{
  CArray<T> Asrt;
  SortValue(Asrt, A);

  int sz = A.m_nData;

  T median;
  int medIdx = (sz - 1)/2;

  if (sz%2 == 0)
      median = (T)((Asrt[medIdx] + Asrt[medIdx+1])/2.0f);
  else
      median = Asrt[medIdx];

  return median;
}

// Variance
template <class T>
T Variance(const CArray<T> &A)
{
    T mn = Mean(A);
    CArray<T> cA = A - mn;
    return (Sumsq(cA)/(float)A.m_nData);
}

template <class T>
T StdDev(const CArray<T> &A)
{
    return sqrt(Variance(A));
}

// Norm
template <class T>
T Norm(const CArray<T> &A)
{
  return sqrt(Sumsq(A));
}

// Max
template <class T> void Max(int &maxIdx, T &maxVal, const CArray<T> &A)
{    
    if (A.isEmpty())
    { maxIdx = -1; return; }
    
    maxVal = A.m_data[0];
    maxIdx = 0;
    
    for (int i = 1; i < A.m_nData; i++)
    {
        if (A[i] > maxVal)
        {
            maxVal = A[i];
            maxIdx = i;
        }        
    }    
}

template <class T> void Max(CArrayi &maxSubs, T &maxVal, const CArray<T> &A)
{
    int maxIdx;
    Max(maxIdx, maxVal, A);
    A.ind2sub(maxSubs, maxIdx);    
}

template <class T> int Maxi(const CArray<T> &A)
{
    int maxIdx;
    T maxVal;
    Max(maxIdx, maxVal, A);
    return maxIdx;
}

template <class T> T Max(const CArray<T> &A)
{
    int maxIdx;
    T maxVal;
    Max(maxIdx, maxVal, A);
    return maxVal;
}

/*template <class T>
T Max(const CArray<T> &A)
{
    T mx = A.m_data[0];
    
    for (int i = 0; i < A.m_nData; i++)
        mx = max(A.m_data[i], mx);
    
    return mx;
    }*/

// Min
template <class T> void Min(int &minIdx, T &maxVal, const CArray<T> &A)
{    
    if (A.isEmpty())
    { minIdx = -1; return; }
    
    maxVal = A.m_data[0];
    minIdx = 0;
    
    for (int i = 1; i < A.m_nData; i++)
    {
        if (A[i] < maxVal)
        {
            maxVal = A[i];
            minIdx = i;
        }        
    }    
}

template <class T> int Mini(const CArray<T> &A)
{
    int minIdx;
    T maxVal;
    Min(minIdx, maxVal, A);
    return minIdx;
}

template <class T> T Min(const CArray<T> &A)
{
    int minIdx;
    T maxVal;
    Min(minIdx, maxVal, A);
    return maxVal;
}

 /*template <class T>
T Min(const CArray<T> &A)
{
    T mn = A.m_data[0];
    
    for (int i = 0; i < A.m_nData; i++)
        mn = min(A.m_data[i], mn);
    
    return mn;
    }*/

template <class T> void NormaliseRange(CArray<T> &Adst, const CArray<T> &Asrc, T mn, T mx)
{
    T minA = Min(Asrc);
    T maxA = Max(Asrc);
    
    T rangeA = maxA - minA;
    T range  = mx - mn;

    // TODO: float version that combines the flops
    Adst = Asrc;
    Adst -= minA;
    Adst *= range;
    Adst /= rangeA;
    Adst += mn;
}

template <class T> CArray<T> NormaliseRange(const CArray<T> &Asrc, T mn, T mx)
{
    CArray<T> Adst;
    Normalise(Adst, Asrc, mn, mx);
    return Adst;
}

template <class T> void NormaliseCentred(CArray<T> &Adst, const CArray<T> &Asrc, T mn, T mx)
{
    T minA = Min(Asrc);
    T maxA = Max(Asrc);

    T rangeA = max(-minA, maxA);
    T range  = (mx - mn) / (T)2;
    
    T ctr    = (mx + mn) / (T)2;

    // TODO: float version that combines the flops    
    Adst = Asrc;
    Adst *= range;
    Adst /= rangeA;
    Adst += ctr;
}

template <class T> CArray<T> NormaliseCentred(const CArray<T> &Asrc, T mn, T mx)
{
    CArray<T> Adst;
    NormaliseCentred(Adst, Asrc, mn, mx);
    return Adst;
}

template <class T> void NormaliseUnit(CArray<T> &Adst, const CArray<T> &Asrc)
{
    Adst = Asrc;
    Adst /= Norm(Asrc);
}

template <class T> CArray<T> NormaliseUnit(const CArray<T> &Asrc)
{
    CArray<T> Adst;
    NormaliseUnit(Adst, Asrc);
    return Adst;
}

template <class T> void ClipMinMax(CArray<T> &A, const T mn, const T mx)
{
    T* Aptr = A.m_data;
    for (int i = 0; i < A.nData(); i++)
    {
        *Aptr = min(mx, max(mn, *Aptr));
        Aptr++;
    }
}

// convert to/from std vector

template <class T> 
void Convert(vector<T> &v, const CArray<T> &A)
{
    v.resize(A.m_nData);
    for (int i = 0; i < A.m_nData; i++)
        v[i] = A[i];
}

template <class T>
void Convert(CArray<T> &A, const vector<T> &v)
{
  A.Create(v.size());
  for (int i = 0; i < A.m_nData; i++)
    A.m_data[i] = v[i];
  return;
}
    
// input/output

template <class T>
ostream& operator<<(ostream &os, const CArray<T> &A)
{
    os << A.m_nDims << endl;
    for (int i = 0; i < A.m_nDims; i++)
    {
        os << A.m_dims[i] << " ";
    }
    os << endl;
    for (int i = 0; i < A.m_nData; i++)
    {
        os << A.m_data[i] << " ";
    }
    os << endl;
    return os;
}

template <class T>
istream& operator>>(istream &is, CArray<T> &A)
{
    int nDims;
    is >> nDims;
    int *dim = new int[nDims];
    for (int i = 0; i < nDims; i++)
    {
        is >> dim[i];
    }

    A.Create(dim, nDims);

    for (int i = 0; i < A.m_nData; i++)
    {
        is >> A.m_data[i];
    }

    delete[] dim;
    return is;
}

/*
template <class T>
void Write(const CArray<T> &arr, string filename)
{
    FILE* fptr = fopen(filename.c_str(), "w");
    fprintf(fptr, "%d\n", arr.m_nDims);
    for (int i = 0; i < arr.m_nDims; i++)
        fprintf(fptr, "%d ", arr.m_dims[i]);
    fprintf(fptr, "\n");
    
    fwrite(arr.m_data, sizeof(T), arr.m_nData, fptr);
    fclose(fptr);
}
    
template <class T>
void Read(CArray<T> &arr, string filename)
{
    FILE* fptr = fopen(filename.c_str(), "r");
    int nDims;
    int rval = fscanf(fptr, "%d\n", &nDims);
    int* dims = new int[nDims];

    // fscanf is broken
    //for (int i = 0; i < nDims; i++)
    //    rval = fscanf(fptr, "%d ", &dims[i]);
    
    // use stringstream instead..
    char buff[200];
    char *dum;
    dum = fgets(buff, 200, fptr);

    stringstream ss(buff);    
    for (int i = 0; i < nDims; i++)
        ss >> dims[i];
                
    arr.Create(dims, nDims);
    delete[] dims;

    rval = fread(arr.m_data, sizeof(T), arr.m_nData, fptr);
    fclose(fptr);
}
*/

template <class T>
void Write(const CArray<T> &arr, string filename)
{
    FILE* fptr = fopen(filename.c_str(), "wb");

    fwrite(&arr.m_nDims, sizeof(int), 1,           fptr);
    fwrite(arr.m_dims,   sizeof(int), arr.m_nDims, fptr);
    fwrite(arr.m_data,   sizeof(T),   arr.m_nData, fptr);

    fclose(fptr);
}

template <class T>
void Read(CArray<T> &arr, string filename)
{
    FILE* fptr = fopen(filename.c_str(), "rb");
    
    int nDims;
    int rval = fread(&nDims, sizeof(int), 1, fptr);
    
    int *dims = new int[nDims];
    rval = fread(dims, sizeof(int), nDims, fptr);
    
    arr.Create(dims, nDims);
    delete[] dims;

    rval = fread(arr.m_data, sizeof(T), arr.m_nData, fptr);
    fclose(fptr);
}
