// ImageBase.h

#pragma once

#include "Array.h"

//#ifndef _USE_MATH_DEFINES
//#define _USE_MATH_DEFINES
//#endif
#include <climits>
#include <cfloat>

#include "math.h"

#include <iostream>
using namespace std;

template <class T> 
class CImage : public CArray<T>
{
 public:
  CImage();
  ~CImage();

  CImage(int nRows, int nCols, int nBands);
  
  CImage(const CImage<T> &M);
  CImage(const CArray<T> &A);

  CImage& operator=(const CImage<T> &M);
  CImage& operator=(const CArray<T> &A);
  CImage& operator=(const T t);

  void Create(int nRows, int nCols, int nBands); 
  void Create(const CArray<int> &size);
    
  T& operator()(int row, int col, int band);
  T& pixelReference(int row, int col, int band) const; // const version of the above
  const T operator()(int row, int col, int band) const;
  const T pixelCopy(int row, int col, int band) const;

  int rowStride(void) const;
  int colStride(void) const;
  int bandStride(void) const;

  CArrayi size(void) const;
  int nRows(void) const;
  int nCols(void) const;
  int nPixels(void) const;
  int nBands(void) const;
};

typedef CImage<float>         CImagef;
typedef CImage<int>           CImagei;
typedef CImage<unsigned char> CImageu;

//
// Implementation
//

template <class T>
CImage<T>::CImage()
{}

template <class T>
CImage<T>::~CImage()
{}

template <class T>
CImage<T>::CImage(int nRows, int nCols, int nBands)
{
  Create(nRows, nCols, nBands);
}

template <class T>
CImage<T>::CImage(const CImage<T> &M)
{
  //this->setDataNull(); // not needed as CArray<T>::CArray() is called
  this->operator=(M);
}

template <class T>
CImage<T>::CImage(const CArray<T> &A)
{
    //this->setDataNull();
  this->operator=(A);
  //this->Make3D();
}

template <class T>
CImage<T>& CImage<T>::operator=(const CImage<T> &M)
{
  CArray<T> *Aptr = (CArray<T> *)(&M);
  this->operator=(*Aptr);
  this->Make3D();
  return *this;
}

template <class T>
CImage<T>& CImage<T>::operator=(const CArray<T> &A)
{
  ((CArray<T> *)this)->operator=(A);
  this->Make3D();
  return *this;
}

template <class T>
CImage<T>& CImage<T>::operator=(const T t)
{
  ((CArray<T> *)this)->operator=(t);
  this->Make3D();
  return *this;
}



template <class T>
void CImage<T>::Create(int nRows, int nCols, int nBands)
{  
  //cerr << "CImage<T>::Create()" << endl;
  int nDims = 3;
  int dims[3];
  dims[0] = nRows;
  dims[1] = nCols;
  dims[2] = nBands;

  ((CArray<T> *)this)->Create(dims, nDims);
}

template <class T>
void CImage<T>::Create(const CArray<int> &size)
{
  Create(size[0], size[1], size[2]);
}

template <class T>
T& CImage<T>::operator()(int row, int col, int band)
{
  // return pixelReference(row, col, band);

  int index = row * this->m_stride[0] + col * this->m_stride[1] + band * this->m_stride[2];
  return this->m_data[index];
}

template <class T>
T& CImage<T>::pixelReference(int row, int col, int band) const
{
  /*int sub[3];
  sub[0] = row;
  sub[1] = col;
  sub[2] = band;
  int index = this->sub2ind(sub);
  return this->m_data[index];*/

  //int index = band + col * nBands() + row * nCols() * nBands();

  int index = row * this->m_stride[0] + col * this->m_stride[1] + band * this->m_stride[2];
  return this->m_data[index];
}

template <class T>
const T CImage<T>::operator()(int row, int col, int band) const
{
  /*int sub[2];
    sub[0] = row;
    sub[1] = col;
    sub[2] = band;
    int index = this->sub2ind(sub);
    return this->m_data[index];*/
  
    int index = row * this->m_stride[0] + col * this->m_stride[1] + band * this->m_stride[2];
    return this->m_data[index];
}

template <class T>
const T CImage<T>::pixelCopy(int row, int col, int band) const
{
  int index = row * this->m_stride[0] + col * this->m_stride[1] + band * this->m_stride[2];
  return this->m_data[index];
}

template <class T>
int CImage<T>::rowStride(void) const
{
  return this->m_stride[0];
  //return(1);
}

template <class T>
int CImage<T>::colStride(void) const
{
  return this->m_stride[1];
  //return(nRows);
}

template <class T>
int CImage<T>::bandStride(void) const
{
  return this->m_stride[2];
}

template <class T>
CArrayi CImage<T>::size(void) const
{
    CArrayi sz;
    
    if (this->isEmpty())
    {
        sz.Create(3);
        sz = 0;
        return sz;
    }
    
    return ((CArray<T>*)this)->size();
}

template <class T>
int CImage<T>::nRows(void) const
{
    if (this->isEmpty())
        return 0;

    return this->m_dims[0];
}

template <class T>
int CImage<T>::nCols(void) const
{
    if (this->isEmpty())
        return 0;

    return this->m_dims[1];
}

template <class T>
int CImage<T>::nPixels(void) const
{
    if (this->isEmpty())
        return 0;
    
    return this->m_dims[0] * this->m_dims[1];
}

template <class T>
int CImage<T>::nBands(void) const
{
    if (this->isEmpty())
        return 0;
    
    return this->m_dims[2];
}
