// ImageFun.h

#pragma once

#include <assert.h>

// unary image operations

template <class T> void Transpose(CImage<T> &im, const CImage<T> &im1); // transpose
template <class T> CImage<T> operator~(const CImage<T> &im1);

// convolve, copy and select

template <class T, class S> void Convolve(T* dstptr, T* srcptr, int nPixels, int nBands, const CArray<S> &kernel); 
template <class T, class S> void ConvolveRows(CImage<T> &im, const CImage<T> &im1, const CArray<S> &kernel);
template <class T, class S> void ConvolveRows(CImage<T> &im, const CArray<S> &kernel); // in place
template <class T, class S> void ConvolveCols(CImage<T> &im, const CImage<T> &im1, const CArray<S> &kernel);
template <class T, class S> void ConvolveCols(CImage<T> &im, const CArray<S> &kernel); 
template <class T, class S> void ConvolveSeparable(CImage<T> &im, const CImage<T> &im1, const CArray<S> &kernel);
template <class T, class S> void ConvolveSeparable(CImage<T> &im, const CArray<S> &kernel);
template <class T> void ConvolveGaussian(CImage<T> &cim, const CImage<T> &im, float sigma);
template <class T> void ConvolveGaussian(CImage<T> &im, float sigma);

template <class T> void Copy(CImage<T> &imDst, int rowDst, int colDst, 
                             const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2);
template <class T> void Copy(CImage<T> &imDst, int rowDst, int colDst, const CImage<T> &imSrc);
template <class T> void Copy(CImage<T> &imDst, const CImage<T> &imSrc);

template <class T> void Add(CImage<T> &imDst, int rowDst, int colDst, 
                            const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2);
template <class T> void Add(CImage<T> &imDst, int rowDst, int colDst, const CImage<T> &imSrc);
template <class T> void Max(CImage<T> &imDst, int rowDst, int colDst, 
                            const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2);
template <class T> void Max(CImage<T> &imDst, int rowDst, int colDst, const CImage<T> &imSrc);

template <class T> void Select(CImage<T> &imDst, const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2);
template <class T> CImage<T> Select(const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2);
template <class T> void SelectRow(CImage<T> &imDst, const CImage<T> &imSrc, int row);
template <class T> CImage<T> SelectRow(const CImage<T> &imSrc, int row);
template <class T> void SelectCol(CImage<T> &imDst, const CImage<T> &imSrc, int col);
template <class T> CImage<T> SelectCol(const CImage<T> &imSrc, int col);

template <class T> T BilinearInterpolate(const CImage<T> &im, float row, float col, int band, bool &isValid=NUseful::dumBool);

// utility

template <class T> void ConvertToGrey(CImage<T>& gim, const CImage<T>& im);
template <class T> CImage<T> ConvertToGrey(const CImage<T>& im);
template <class T> void ConvertToColour(CImage<T>& cim, const CImage<T>& im);
template <class T> CImage<T> ConvertToColour(const CImage<T>& im);
template <class T> void TransformColour(CImage<T> &tim, const CImage<T> &im, const CArray<T> &M);

template <class T> void NormaliseImage(CImage<T>& nim, const CImage<T>& im);
template <class T> CImage<T> NormaliseImage(const CImage<T>& im);
template <class T> void NormaliseImageCentred(CImage<T>& nim, const CImage<T>& im);
template <class T> CImage<T> NormaliseImageCentred(const CImage<T>& im);

template <class T> void DownSample2_Nearest(CImage<T> &sim, const CImage<T> &im);
template <class T> void DownSample2_Nearest(CImage<T> &im);
template <class T> void DownSample2_2x2(CImage<T> &sim, const CImage<T> &im);
template <class T> void DownSample2_2x2(CImage<T> &im);
template <class T> void DownSample2_3x3(CImage<T> &sim, const CImage<T> &im);
template <class T> void DownSample2_3x3(CImage<T> &im);
template <class T> void DownSample2ToMinSize(CImage<T> &sim, const CImage<T> &im, int minSize);

template <class T> void UpSample2_Bilinear(CImage<T> &uim, const CImage<T> &im); 
template <class T> void UpSample2_Bilinear(CImage<T> &im);
template <class T> void UpSample2_Bilinear(CImage<T> &uim, const CImage<T> &im, int num);
template <class T> void UpSample2_Bilinear(CImage<T> &im, int num);
template <class T> void UpSample2_Nearest(CImage<T> &uim, const CImage<T> &im);
template <class T> void UpSample2_Nearest(CImage<T> &im);
template <class T> void UpSample2_Nearest(CImage<T> &uim, const CImage<T> &im, int num);

template <class T> void GetBand(CImage<T> &band, const CImage<T> &im, int nBand);
template <class T> void SetBand(CImage<T> &im, const CImage<T> &band, int nBand);
template <class T> void SetBand(CImage<T> &im, T val, int nBand);

template <class T> void Rotate90cw(CImage<T> &dst, const CImage<T> &src);
template <class T> void Rotate90acw(CImage<T> &dst, const CImage<T> &src);
template <class T> void Rotate180(CImage<T> &dst, const CImage<T> &src);

template <class T> void ScaleBilinear(CImage<T> &dst, const CImage<T> &src, float scale);
template <class T> void ScaleBilinearNoCreate(CImage<T> &dst, const CImage<T> &src, float scale);

template <class T> void ExtendZeroOrder(CImage<T> &extIm, const CImage<T> &im, int nPixels);
template <class T> void ExtendZeroOrderRB(CImage<T> &extIm, const CImage<T> &im, int nPixels);

template <class T> CArray<T> PixelColour(const CImage<T> &im, int row, int col);
template <class T> void DrawRect(CImage<T> &im, int row0, int col0, int row1, int col1, CArray<T> &colour);

//
// Implementation
//

// unary image operations

template <class T>
void Transpose(CImage<T> &im, const CImage<T> &im1)
{
    // TODO: exploit caching e.g. cache oblivious matrix transpose

    int nBands = im1.nBands();
    im.Create(im1.nCols(), im1.nRows(), nBands);
  
    T* dstptr = im.m_data;

    T*  srcColptr    = im1.m_data;
    int srcRowStride = im1.rowStride();
    int srcColStride = im1.colStride();

    for (int r = 0; r < im.nRows(); r++)
    {
        T* srcptr = srcColptr;
      
        for (int c = 0; c < im.nCols(); c++)
	{	  
            memcpy(dstptr, srcptr, nBands * sizeof(T));
            srcptr += srcRowStride;
            dstptr += nBands;
	}

        srcColptr += srcColStride;
    }
}

template <class T>
CImage<T> operator~(const CImage<T> &im1)
{
    CImage<T> im;
    Transpose(im, im1);
    return im;   
}

// convolve, copy and select

template <class T, class S> void Convolve(T* dstptr, T* srcptr, int nPixels, int nBands, const CArray<S> &kernel)
{    
    int nData = nPixels * nBands;
  
    int ksize = kernel.m_nData;
    S* kptr0  = kernel.m_data;
    
    // make sure kernel length is odd
    CArray<S> kernel2;
    if (ksize % 2 == 0)
    {
        kernel2.Create(ksize + 1);
        for (int i = 0; i < kernel.m_nData; i++)
            kernel2[i] = kernel[i];
        kptr0 = kernel2.m_data;
        ksize++;
    }
  
    int length = (ksize - 1)/2;
    int nPixelsBuffer = nPixels + 2 * length;
    int nDataBuffer = nPixelsBuffer * nBands;

    // extend data outside range
    CArray<T> buffer(nDataBuffer);
    T *bufferptr = buffer.m_data;
  
    // extend left
    for (int i = 0; i < length; i++)
    {
        memcpy(bufferptr, srcptr, nBands * sizeof(T));
        bufferptr += nBands;
    }
	
    // copy centre
    memcpy(bufferptr, srcptr, nData * sizeof(T));
    bufferptr += nData;
    srcptr += (nData - nBands);

    // extend right
    for (int i = 0; i < length; i++)
    {
        memcpy(bufferptr, srcptr, nBands * sizeof(T));
        bufferptr += nBands;
    }

    memset(dstptr, 0, nData * sizeof(T)); // needed for in-place use 
       
    // Convolve buffer with kernel    
    S *kptr;
    S *kptr1      = kptr0 + ksize - 1;
    T *bufferptr0 = buffer.m_data;

    for (int i = 0; i < nPixels; i++)
    {
        kptr      = kptr1;
        bufferptr = bufferptr0;
      
        for (int k = 0; k < ksize; k++)
        {  
            T kval = (T)(*kptr);
	  
            for (int b = 0; b < nBands; b++)
	    {
                *dstptr += *bufferptr * kval;
                dstptr++;
                bufferptr++;
	    }

            dstptr -= nBands;
            kptr--;
        }

        dstptr += nBands;
        bufferptr0 += nBands;
    }
  
}

template <class T, class S> void ConvolveRows(CImage<T> &im, const CImage<T> &im1, const CArray<S> &kernel)
{
    im.Create(im1.size());
        
    T* srcptr = im1.m_data;
    T* dstptr = im.m_data;

    for (int r = 0; r < im.nRows(); r++)
    {
        Convolve(dstptr, srcptr, im.nCols(), im.nBands(), kernel);      
        srcptr += im1.rowStride();
        dstptr += im.rowStride();        
    }
        

    // omp
    /*
    T* srcptr0 = im1.m_data;
    T* dstptr0 = im.m_data;

    #pragma omp parallel for
    for (int r = 0; r < im.nRows(); r++)
    {
        T* srcptr = srcptr0 + im1.rowStride() * r;
        T* dstptr = dstptr0 + im.rowStride() * r;
        Convolve(dstptr, srcptr, im.nCols(), im.nBands(), kernel);      
    }
    */

}

template <class T, class S> void ConvolveRows(CImage<T> &im, const CArray<S> &kernel)
{  
    
    T* rowptr = im.m_data;

    for (int r = 0; r < im.nRows(); r++)
    {
        Convolve(rowptr, rowptr, im.nCols(), im.nBands(), kernel);
        rowptr += im.rowStride();
    }
    

    // omp
    /*
    T* rowptr0 = im.m_data;

    #pragma omp parallel for
    for (int r = 0; r < im.nRows(); r++)
    {
        T* rowptr = rowptr0 + im.rowStride() * r;
        Convolve(rowptr, rowptr, im.nCols(), im.nBands(), kernel);
    }
    */
}

template <class T, class S> void ConvolveCols(CImage<T> &im, const CImage<T> &im1, const CArray<S> &kernel)
{
    CImage<T> imt;
    Transpose(imt, im1);
    ConvolveRows(imt, kernel);  
    Transpose(im, imt);
}

template <class T, class S> void ConvolveCols(CImage<T> &im, const CArray<S> &kernel)
{
    CImage<T> imt;
    Transpose(imt, im);
    ConvolveRows(imt, kernel);
    Transpose(im, imt);
}

template <class T, class S> void ConvolveSeparable(CImage<T> &im, const CImage<T> &im1, const CArray<S> &kernel)
{
    ConvolveRows(im, im1, kernel);
    ConvolveCols(im, kernel);
}

template <class T, class S> void ConvolveSeparable(CImage<T> &im, const CArray<S> &kernel)
{
    ConvolveRows(im, kernel);
    ConvolveCols(im, kernel);
}

template <class T> void ConvolveGaussian(CImage<T> &cim, const CImage<T> &im, float sigma)
{
    int length = (int)floor(3 * sigma + 0.5);
  
    if (length < 1)
    {
        cim = im;
        return;
    }

    CArrayf kernel = GaussKernel(sigma, length); 

    ConvolveSeparable(cim, im, kernel);
}

template <class T> void ConvolveGaussian(CImage<T> &im, float sigma)
{
    int length = (int)floor(3 * sigma + 0.5);
    
    if (length < 1)
        return;
 
    CArrayf kernel = GaussKernel(sigma, length);
 
    ConvolveSeparable(im, kernel);
}

template <class T> void Copy(CImage<T> &imDst, int rowDst, int colDst, 
                             const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2)
{
    
    if (imDst.colStride() == imSrc.colStride())
    {
        // memcpy rows

        int nRowsCopy  = min(imDst.nRows() - rowDst, rowSrc2 - rowSrc1 + 1);
        int nColsCopy  = min(imDst.nCols() - colDst, colSrc2 - colSrc1 + 1);
    
        rowSrc2 = rowSrc1 + nRowsCopy - 1;
        colSrc2 = colSrc1 + nColsCopy - 1;
    
        //T* srcRowPtr = &imSrc(rowSrc1, colSrc1, 0);
        T* srcRowPtr = &imSrc.pixelReference(rowSrc1, colSrc1, 0);
        T* dstRowPtr = &imDst(rowDst, colDst, 0);

        int nBytesPerRow = nColsCopy * imDst.colStride() * sizeof(T);

        for (int r = 0; r < nRowsCopy; r++)
        {
            memcpy(dstRowPtr, srcRowPtr, nBytesPerRow);
            srcRowPtr += imSrc.rowStride();
            dstRowPtr += imDst.rowStride();
        }
    }
    else 
    {
        // imDst.colStride() != imSrc.colStride()
        CArrayi dstStart(3);
        dstStart[0] = rowDst;
        dstStart[1] = colDst;
        dstStart[2] = 0;
        
        CArrayi srcStart(3);
        srcStart[0] = rowSrc1;
        srcStart[1] = colSrc1;
        srcStart[2] = 0;
        
        CArrayi srcEnd(3);
        srcEnd[0] = rowSrc2;
        srcEnd[1] = colSrc2;
        srcEnd[2] = imSrc.nBands() - 1;
        
        Copy(imDst, dstStart, imSrc, srcStart, srcEnd);
    }
}

template <class T> void Copy(CImage<T> &imDst, int rowDst, int colDst, const CImage<T> &imSrc)
{
    Copy(imDst, rowDst, colDst, imSrc, 0, 0, imSrc.nRows() - 1, imSrc.nCols() - 1);
}

template <class T> void Copy(CImage<T> &imDst, const CImage<T> &imSrc)
{
    Copy(imDst, 0, 0, imSrc, 0, 0, imSrc.nRows() - 1, imSrc.nCols() - 1);
}

template <class T> void Add(CImage<T> &imDst, int rowDst, int colDst, 
                             const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2)
{  
    int nRowsAdd  = min(imDst.nRows() - rowDst, rowSrc2 - rowSrc1 + 1);
    int nColsAdd  = min(imDst.nCols() - colDst, colSrc2 - colSrc1 + 1);
    int nBands    = imSrc.nBands();

    rowSrc2 = rowSrc1 + nRowsAdd - 1;
    colSrc2 = colSrc1 + nColsAdd - 1;
    
    T* srcRowPtr = &imSrc.pixelReference(rowSrc1, colSrc1, 0);
    T* dstRowPtr = &imDst(rowDst, colDst, 0);

    for (int r = 0; r < nRowsAdd; r++)
    {
        T* srcPtr = srcRowPtr;
        T* dstPtr = dstRowPtr;
        for (int c = 0; c < nColsAdd; c++)
        {
            for (int b = 0; b < nBands; b++)
                *dstPtr++ += *srcPtr++;
        }
        srcRowPtr += imSrc.rowStride();
        dstRowPtr += imDst.rowStride();
    }
}

template <class T> void Add(CImage<T> &imDst, int rowDst, int colDst, const CImage<T> &imSrc)
{
    Add(imDst, rowDst, colDst, imSrc, 0, 0, imSrc.nRows() - 1, imSrc.nCols() - 1);
}

template <class T> void Max(CImage<T> &imDst, int rowDst, int colDst, 
                             const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2)
{  
    int nRowsAdd  = min(imDst.nRows() - rowDst, rowSrc2 - rowSrc1 + 1);
    int nColsAdd  = min(imDst.nCols() - colDst, colSrc2 - colSrc1 + 1);
    int nBands    = imSrc.nBands();

    rowSrc2 = rowSrc1 + nRowsAdd - 1;
    colSrc2 = colSrc1 + nColsAdd - 1;
    
    T* srcRowPtr = &imSrc.pixelReference(rowSrc1, colSrc1, 0);
    T* dstRowPtr = &imDst(rowDst, colDst, 0);

    for (int r = 0; r < nRowsAdd; r++)
    {
        T* srcPtr = srcRowPtr;
        T* dstPtr = dstRowPtr;
        for (int c = 0; c < nColsAdd; c++)
        {
            for (int b = 0; b < nBands; b++)
            {
                float &dst = *dstPtr;
                dst = max(dst, *srcPtr++);
                dstPtr++;
            }
        }
        srcRowPtr += imSrc.rowStride();
        dstRowPtr += imDst.rowStride();
    }
}

template <class T> void Max(CImage<T> &imDst, int rowDst, int colDst, const CImage<T> &imSrc)
{
    Max(imDst, rowDst, colDst, imSrc, 0, 0, imSrc.nRows() - 1, imSrc.nCols() - 1);
}


template <class T> void Select(CImage<T> &imDst, const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2)
{
    int nRowsCopy = rowSrc2 - rowSrc1 + 1;
    int nColsCopy = colSrc2 - colSrc1 + 1;

    imDst.Create(nRowsCopy, nColsCopy, imSrc.nBands());
    
    T* srcRowPtr = &imSrc.pixelReference(rowSrc1, colSrc1, 0);
    T* dstRowPtr = imDst.m_data;

    int nBytesPerRow = nColsCopy * imSrc.colStride() * sizeof(T);

    for (int r = 0; r < nRowsCopy; r++)
    {
        memcpy(dstRowPtr, srcRowPtr, nBytesPerRow);
        srcRowPtr += imSrc.rowStride();
        dstRowPtr += imDst.rowStride();
    }
}

template <class T> CImage<T> Select(const CImage<T> &imSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2)
{
    CImage<T> imDst;
    Select(imDst, imSrc, rowSrc1, colSrc1, rowSrc2, colSrc2);
    return imDst;
}

template <class T> void SelectRow(CImage<T> &imDst, const CImage<T> &imSrc, int row)
{
    Select(imDst, imSrc, row, 0, row, imSrc.nCols() - 1);
}

template <class T> CImage<T> SelectRow(const CImage<T> &imSrc, int row)
{
    CImage<T> imDst;
    SelectRow(imDst, imSrc, row);
    return imDst;
}

template <class T> void SelectCol(CImage<T> &imDst, const CImage<T> &imSrc, int col)
{
    Select(imDst, imSrc, 0, col, imSrc.nRows() - 1, col);
}

template <class T> CImage<T> SelectCol(const CImage<T> &imSrc, int col)
{
    CImage<T> imDst;
    SelectCol(imDst, imSrc, col);
    return imDst;
}

template <class T> T BilinearInterpolate(const CImage<T> &im, float row, float col, int band, bool &isValid)
{ 
    isValid = true;
    T I = (T)0;
    
    if (!(row >= 0 && row < im.nRows() - 1 && col >= 0 && col < im.nCols() - 1))
    {
        isValid = false;
        return I;
    }

    int ri = (int)row;
    int ci = (int)col;
    float rf = row - (float)ri;
    float cf = col - (float)ci;
    
    T I00 = im(ri,     ci,     band);
    T I01 = im(ri,     ci + 1, band);
    T I10 = im(ri + 1, ci,     band);
    T I11 = im(ri + 1, ci + 1, band);
    
    I  = (1.0f - rf) * (1.0f - cf) * I00 +
         (1.0f - rf) *         cf  * I01 + 
                 rf  * (1.0f - cf) * I10 + 
                 rf  *         cf  * I11;
    
    return I;
}
  
// utility

template <class T> void ConvertToGrey(CImage<T>& gim, const CImage<T>& im)
{
    gim.Create(im.nRows(), im.nCols(), 1);

    T* imptr  = im.m_data;
    T* gimptr = gim.m_data;

    int nPixels = im.nRows() * im.nCols();

    for (int i = 0; i < nPixels; i++)
    {
        float sum = 0.0f;
      
        for (int b = 0; b < im.nBands(); b++)
	{
            sum += (float)*imptr;
            imptr++;
	}
      
        *gimptr = (T)(sum / (float)im.nBands());
    
        gimptr++;
    }
}
      
template <class T> CImage<T> ConvertToGrey(const CImage<T> &im)
{
    CImage<T> gim;
    ConvertToGrey(gim, im);
    return gim;
}


template <class T> void ConvertToColour(CImage<T> &cim, const CImage<T> &im)
{
    int nBands = 3;
    cim.Create(im.nRows(), im.nCols(), nBands);

    T* imptr  = im.m_data;
    T* cimptr = cim.m_data;

    for (int i = 0; i < im.m_nData; i++)
    {
        for (int b = 0; b < nBands; b++)
	{
            *cimptr = *imptr;
            cimptr++;
	}
        imptr++;
    }
}

template <class T> CImage<T> ConvertToColour(const CImage<T> &im)
{
    CImage<T> cim;
    ConvertToColour(cim, im);
    return cim;
}

template <class T> void TransformColour(CImage<T> &tim, const CImage<T> &im, const CArray<T> &M)
{
    assert(im.nBands() == 3);

    tim.Create(im.size());
//    T m00 = M(0, 0); T m01 = M(0, 1); T m02 = M(0, 2);
//    T m10 = M(1, 0); T m11 = M(1, 1); T m12 = M(1, 2);
//    T m20 = M(2, 0); T m21 = M(2, 1); T m22 = M(2, 2);

    T m00 = M[0]; T m01 = M[1]; T m02 = M[2];
    T m10 = M[3]; T m11 = M[4]; T m12 = M[5];
    T m20 = M[6]; T m21 = M[7]; T m22 = M[8];

    T* imptr = im.m_data;
    T* timptr = tim.m_data;

    for (int i = 0; i < im.nPixels(); i++)
    {
        T r = *imptr++;
        T g = *imptr++;
        T b = *imptr++;
        
        T rt = m00 * r + m01 * g + m02 * b;
        T gt = m10 * r + m11 * g + m12 * b;
        T bt = m20 * r + m21 * g + m22 * b;

        *timptr++ = rt;
        *timptr++ = gt;
        *timptr++ = bt;
    }        
}

template <class T> void NormaliseImage(CImage<T>& nim, const CImage<T>& im)
{
    NormaliseRange(nim, im, (T)0, (T)255);
}

template <class T> CImage<T> NormaliseImage(const CImage<T>& im)
{
    CImage<T> nim;
    NormaliseImage(nim, im);
    return nim;
}

template <class T> void NormaliseImageCentred(CImage<T>& nim, const CImage<T>& im)
{
    NormaliseCentred(nim, im, (T)0, (T)255);
}

template <class T> CImage<T> NormaliseImageCentred(const CImage<T>& im)
{
    CImage<T> nim;
    NormaliseImageCentred(nim, im);
    return nim;
}

template <class T> void DownSample2_Nearest(CImage<T> &sim, const CImage<T> &im)
// Subsample factor 2 using nearest pixel resampling
{
    int nRows  = im.nRows();
    int nCols  = im.nCols();
    int nBands = im.nBands();
    int nBands2 = nBands * 2;
 
    // round up dims divide by 2
    int nRows2 = im.nRows() / 2; 
    if (im.nRows() % 2 == 1)
        nRows2++;
    int nCols2 = im.nCols() / 2;
    if (im.nCols() % 2 == 1)
        nCols2++;

    sim.Create(nRows2, nCols2, nBands);

    T *simptr   = sim.m_data;
    T *imptr    = im.m_data;
    T *imrowptr = im.m_data;
    int imRowStride = im.rowStride();
    int imRowStride2 = imRowStride * 2;

    for (int i = 0; i < nRows2; i++)
    {
        imptr = imrowptr;

        for (int j = 0; j < nCols2; j++)
        {         
	  memcpy(simptr, imptr, nBands * sizeof(T));
	  simptr += nBands;
	  imptr += nBands2;
        }// end for j

        imrowptr += imRowStride2;
    }// end for i
}

template <class T> void DownSample2_Nearest(CImage<T> &im)
{
    CImage<T> sim;
    DownSample2_Nearest(sim, im);
    Swap(sim, im);
}

template <class T> void DownSample2_2x2(CImage<T> &sim, const CImage<T> &im)
// Subsample image using 2x2 block average
{	
    int nRows  = im.nRows();
    int nCols  = im.nCols();
    int nBands = im.nBands();
    int nRows2 = (int)floor(nRows / 2.0f); // Force even size
    int nCols2 = (int)floor(nCols / 2.0f);
 
    sim.Create(nRows2, nCols2, nBands);

    T *simptr   = sim.m_data;
    T *imptr    = im.m_data;
    T *imrowptr = im.m_data;
    int imRowStride = im.rowStride();
    int imRowStride2 = imRowStride * 2;

    for (int i = 0; i < nRows2; i++)
    {
        imptr = imrowptr;

        for (int j = 0; j < nCols2; j++)
        {
            for (int b = 0; b < nBands; b++)
            {
                T *imptrtmp;
                imptrtmp = imptr;
                T avg = 0;
                
                avg += *imptrtmp;
                imptrtmp += nBands;
                avg += *imptrtmp;
                imptrtmp += imRowStride;
                avg += *imptrtmp;
                imptrtmp -= nBands;
                avg += *imptrtmp;
                
                *simptr = avg / 4.0f;
                
                simptr++;
                imptr++;
            }
            
            imptr += nBands;

        }// end for j

        imrowptr += imRowStride2;
    }// end for i

}

template <class T> void DownSample2_2x2(CImage<T> &im)
{
    CImage<T> sim;
    DownSample2_2x2(sim, im);
    Swap(sim, im);
}

template <class T> void DownSample2_3x3(CImage<T> &sim, const CImage<T> &im)
// Subsample image using 3x3 kernel
{	
    // resampling kernel is
    // 0.25 0.5 0.25
    // 0.5  1.0 0.5
    // 0.25 0.5 0.25 
    // divided by 4

    // round up dims divide by 2
    int imRows2 = im.nRows() / 2; 
    if (im.nRows() % 2 == 1)
        imRows2++;
    int imCols2 = im.nCols() / 2;
    if (im.nCols() % 2 == 1)
        imCols2++;
   
    //sim.Create(im.nRows() / 2, im.nCols() / 2, im.nBands());
    sim.Create(imRows2, imCols2, im.nBands());
    
    // extend image by 1 pixel outside it's range
    CImage<T> extIm;
    ExtendZeroOrder(extIm, im, 1);
    
    T *simptr   = sim.m_data;
    T *imptr    = extIm.m_data;
    T *imrowptr = imptr;

    int imColStride = extIm.colStride();
    int imRowStride = extIm.rowStride();
    int nBands      = extIm.nBands();

    for (int r = 0; r < sim.nRows(); r++)
    {
        imptr = imrowptr;
        
        for (int c = 0; c < sim.nCols(); c++)
        {
            for (int b = 0; b < sim.nBands(); b++)
            {
                /*
                 *simptr += *imptr * 0.0625;
                imptr += imColStride;
                *simptr += *imptr * 0.125;
                imptr += imColStride;
                *simptr += *imptr * 0.0625;
            
                imptr += imRowStride;
                *simptr += *imptr * 0.125;
                imptr -= imColStride;
                *simptr += *imptr * 0.25;
                imptr -= imColStride;
                *simptr += *imptr * 0.125;
            
                imptr += imRowStride;
                *simptr += *imptr * 0.0625;
                imptr += imColStride;
                *simptr += *imptr * 0.125;
                imptr += imColStride;
                *simptr += *imptr * 0.0625;
                */
                
                // integer compatible version

                 *simptr += *imptr;
                imptr += imColStride;
                *simptr += *imptr * (T)2;
                imptr += imColStride;
                *simptr += *imptr;
            
                imptr += imRowStride;
                *simptr += *imptr * (T)2;
                imptr -= imColStride;
                *simptr += *imptr * (T)4;
                imptr -= imColStride;
                *simptr += *imptr * (T)2;
            
                imptr += imRowStride;
                *simptr += *imptr;
                imptr += imColStride;
                *simptr += *imptr * (T)2;
                imptr += imColStride;
                *simptr += *imptr;

                *simptr /= (T)16;
            
                imptr -= 2 * imRowStride;
                imptr -= 2 * imColStride;
                imptr++; // next band    

                simptr++;
            }
            imptr -= nBands;
            imptr += 2 * imColStride;            
        }
        
        imrowptr += 2 * imRowStride;        
    }

}

template <class T> void DownSample2_3x3(CImage<T> &im)
{
    CImage<T> sim;
    DownSample2_3x3(sim, im);
    Swap(im, sim);
}

template <class T> void DownSample2ToMinSize(CImage<T> &im, int minSize)
{    
    while (im.nRows() >= 2 * minSize && im.nCols() >= 2 * minSize)
        DownSample2_3x3(im);
}

template <class T> void UpSample2_Bilinear(CImage<T> &uim, const CImage<T> &im)
{    
    int imRows = im.nRows();
    int imCols = im.nCols();
    int uimRows = 2 * imRows;
    int uimCols = 2 * imCols;
    int nBands = im.nBands();

    uim.Create(uimRows, uimCols, nBands);
    int uimRowStride = uim.rowStride();
    int imRowStride = im.rowStride();

    T* imrowptr = im.m_data;
    T* uimrowptr = uim.m_data;

    for (int i = 0; i < imRows - 1; i++)
    {
        T* imptr = imrowptr;
        T* uimptr = uimrowptr;

        for (int j = 0; j < imCols - 1; j++)
        {

            for (int band = 0; band < nBands; band++)
            {
                T I00 = *imptr;
                imptr += nBands;
                T I01 = *imptr;
                imptr += imRowStride;
                T I11 = *imptr;
                imptr -= nBands; 
                T I10 = *imptr;
                imptr -= imRowStride;
                imptr++;
                
                T a = (I00 + I01) * 0.5f;
                T b = (I00 + I10) * 0.5f;                
                T c = 0.25f * (I00 + I10 + I01 + I11);

                *uimptr = I00;
                uimptr += nBands;
                *uimptr = a;
                uimptr += uimRowStride;
                *uimptr = c;
                uimptr -= nBands;
                *uimptr = b;
                uimptr -= uimRowStride;
                uimptr++;
            }
            uimptr += nBands;
        }
        
        imrowptr += imRowStride;
        uimrowptr += 2 * uimRowStride;
        
    }

    // Zero order extend last col
    for (int i = 0; i < imRows - 1; i++)
    {
        for (int b = 0; b < nBands; b++)
        {
            T I0 = im(i, imCols - 1, b);
            T I1 = im(i + 1, imCols - 1, b);
            T a = (I0 + I1) * 0.5f;

            uim(2 * i, uimCols - 2, b) = I0;
            uim(2 * i, uimCols - 1, b) = I0;
            uim(2 * i + 1, uimCols - 2, b) = a;
            uim(2 * i + 1, uimCols - 1, b) = a;
        }
    }

    // Zero order extend last row
    for (int i = 0; i < imCols - 1; i++)
    {
        for (int b = 0; b < nBands; b++)
        {
            T I0 = im(imRows - 1, i, b);
            T I1 = im(imRows - 1, i + 1, b);
            T a = (I0 + I1) * 0.5f;

            uim(uimRows - 2, 2 * i, b) = I0;
            uim(uimRows - 1, 2 * i, b) = I0;
            uim(uimRows - 2, 2 * i + 1, b) = a;
            uim(uimRows - 1, 2 * i + 1, b) = a;
        }
    }
    
    // Bottom right pixel
    for (int b = 0; b < nBands; b++)
    {
        T I = im(imRows - 1, imCols - 1, b);
        uim(uimRows - 2, uimCols - 2, b) = I;
        uim(uimRows - 1, uimCols - 2, b) = I;
        uim(uimRows - 2, uimCols - 1, b) = I;
        uim(uimRows - 1, uimCols - 1, b) = I;
    }

}          

template <class T> void UpSample2_Bilinear(CImage<T> &im)
{
    CImage<T> uim;
    UpSample2_Bilinear(uim, im);
    Swap(uim, im);
}

template <class T> void UpSample2_Bilinear(CImage<T> &uim, const CImage<T> &im, int num)
{
    //uim = im;
    //UpSample2_Bilinear(uim, num);

    if (num < 1)
    {
        uim = im;
        return;
    }
   
    CImage<T> imt;
    UpSample2_Bilinear(imt, im);

    for (int i = 0; i < num - 1; i++)
        UpSample2_Bilinear(imt);
    
    Swap(uim, imt);
}

template <class T> void UpSample2_Bilinear(CImage<T> &im, int num)
{
    for (int i = 0; i < num; i++)
        UpSample2_Bilinear(im);
}

template <class T> void UpSample2_Nearest(CImage<T> &uim, const CImage<T> &im)
{
    int nBands = im.nBands();

    uim.Create(im.nRows() * 2, im.nCols() * 2, nBands);
    
    T* uimptr = uim.m_data;
    T* imptr = im.m_data;
    for (int i = 0; i < im.nRows(); i++)
    {
        for (int j = 0; j < im.nCols(); j++)
        {
            for (int b = 0; b < nBands; b++)
                *uimptr++ = *imptr++;
            imptr -= nBands;

            for (int b = 0; b < nBands; b++)
                *uimptr++ = *imptr++;
        }
        
        memcpy(uimptr, uimptr - uim.rowStride(), im.nCols() * nBands * 2 * sizeof(T));
        uimptr += uim.rowStride();        
    }
}

template <class T> void UpSample2_Nearest(CImage<T> &im)
{
    CImage<T> uim;
    UpSample2_Nearest(uim, im);
    Swap(uim, im);
}

template <class T> void UpSample2_Nearest(CImage<T> &uim, const CImage<T> &im, int num)
{
    if (num < 1)
    {
        uim = im;
        return;
    }
    
    CImage<T> imt;
    UpSample2_Nearest(imt, im);

    for (int i = 0; i < num - 1; i++)
        UpSample2_Nearest(imt);
    
    Swap(uim, imt);
}


template <class T> void GetBand(CImage<T> &band, const CImage<T> &im, int nBand)
{
    band.Create(im.nRows(), im.nCols(), 1);

    T* bandptr = band.m_data;
    T* imptr   = im.m_data + nBand;

    int bandStride = im.nBands();
    int nPixels = im.nRows() * im.nCols();

    for (int i = 0; i < nPixels; i++)
    {
        *bandptr++ = *imptr;
        imptr += bandStride;
    }
}
    
template <class T> void SetBand(CImage<T> &im, const CImage<T> &band, int nBand)
{
    assert(band.nRows() == im.nRows());
    assert(band.nCols() == im.nCols());

    T* bandptr = band.m_data;
    T* imptr   = im.m_data + nBand;

    int bandStride = im.nBands();
    int nPixels = im.nRows() * im.nCols();

    for (int i = 0; i < nPixels; i++)
    {
        *imptr = *bandptr++;
        imptr += bandStride;
    }
}

template <class T> void SetBand(CImage<T> &im, T val, int nBand)
{
    T* imptr   = im.m_data + nBand;

    int bandStride = im.nBands();

    for (int i = 0; i < im.nPixels(); i++)
    {
        *imptr = val;
        imptr += bandStride;
    }
}

template <class T> void Rotate90cw(CImage<T> &dst, const CImage<T> &src)
{
  dst.Create(src.nCols(), src.nRows(), src.nBands());

  T* srcPtr = src.m_data;
  T* dstPtr = dst.m_data; 

  int srcColStride = src.colStride();
  int dstRowStride = dst.rowStride();
  int dstColStride = dst.colStride();
  const int nBands = dst.nBands();

  dstPtr += dstColStride * (dst.nCols() - 1);

  for (int c = dst.nCols() - 1; c >= 0; c--)
    {
      for (int r = 0; r < dst.nRows(); r++)
	{
	  memcpy(dstPtr, srcPtr, nBands * sizeof(T));
	  srcPtr += srcColStride;
	  dstPtr += dstRowStride;
	}
      dstPtr -= dstRowStride * dst.nRows();
      dstPtr -= dstColStride;
    }  
}

template <class T> void Rotate90acw(CImage<T> &dst, const CImage<T> &src)
{
  dst.Create(src.nCols(), src.nRows(), src.nBands());

  T* srcPtr = src.m_data;
  T* dstPtr = dst.m_data; 

  int srcColStride = src.colStride();
  int dstRowStride = dst.rowStride();
  int dstColStride = dst.colStride();
  const int nBands = dst.nBands();

  dstPtr += dstRowStride * (dst.nRows() - 1);
  
  for (int c = 0; c < dst.nCols(); c++)
    {
      for (int r = dst.nRows() - 1; r >= 0; r--)
	{
	  memcpy(dstPtr, srcPtr, nBands * sizeof(T));
	  srcPtr += srcColStride;
	  dstPtr -= dstRowStride;
	}
      dstPtr += dstColStride;
      dstPtr += dstRowStride * dst.nRows();
    }  
}

template <class T> void Rotate180(CImage<T> &dst, const CImage<T> &src)
{
  dst.Create(src.size());
  
  T* srcPtr = src.m_data;
  T* dstPtr = dst.m_data;
  
  int dstColStride = dst.colStride();
  int srcColStride = src.colStride();
  const int nBands = src.nBands();

  dstPtr += dst.nData();
  dstPtr -= dstColStride;

  int nPixels = src.nRows() * src.nCols();

  for (int i = 0; i < nPixels; i++)
    {
      memcpy(dstPtr, srcPtr, nBands * sizeof(T));
      srcPtr += srcColStride;
      dstPtr -= dstColStride;
    }
}


template <class T> void ScaleBilinear(CImage<T> &dst, const CImage<T> &src, float scale)
{
    int dstRows = (int)(ceil(src.nRows() * scale));
    int dstCols = (int)(ceil(src.nCols() * scale));

    dst.Create(dstRows, dstCols, src.nBands());

    ScaleBilinearNoCreate(dst, src, scale);    
}

template <class T> void ScaleBilinearNoCreate(CImage<T> &dst, const CImage<T> &src, float scale)
{
    assert(src.nBands() == dst.nBands());

    float iscale = 1.0f / scale;
    
    for (int rdst = 0; rdst < dst.nRows(); rdst++)
    {
        for (int cdst = 0; cdst < dst.nCols(); cdst++)
        {
            float r = (float)rdst * iscale;
            float c = (float)cdst * iscale;
            
            int ri = (int)r;
            int ci = (int)c;
            
            int ri2 = min(src.nRows() - 1, ri + 1); // zero-order extend last row/col 
            int ci2 = min(src.nCols() - 1, ci + 1); 

            if ((ri < src.nRows()) && (ci < src.nCols()))
            {
                float rf = r - ri;
                float cf = c - ci;
                
                //printf("ri ci rf cf = %d %d %f %f\n", ri, ci, rf, cf);
                //printf("rdst cdst = %d %d\n", rdst, cdst);

                for (int b = 0; b < dst.nBands(); b++)
                {
                    float I00 = src(ri, ci, b);
                    float I01 = src(ri, ci2, b);
                    float I10 = src(ri2, ci, b);
                    float I11 = src(ri2, ci2, b);
                    
                    float I = (1 - rf) * (1 - cf) * I00 + 
                        (1 - rf) * cf * I01 + 
                        rf * (1 - cf) * I10 + 
                        rf * cf * I11;

                    dst(rdst, cdst, b) = I;
                }// end for b                
            }// end if in src image
            

        }// end for cdst
    }// end for rdst
    
}

template <class T> void ExtendZeroOrder(CImage<T> &extIm, const CImage<T> &im, int nPixels)
{
    extIm.Create(im.nRows() + 2 * nPixels, im.nCols() + 2 * nPixels, im.nBands());
    
    Copy(extIm, nPixels, nPixels, im);
    
    for (int i = 0; i < nPixels; i++)
    {
        Copy(extIm, nPixels, i, im, 0, 0, im.nRows() - 1, 0); // left
        Copy(extIm, nPixels, extIm.nCols() - i - 1, im, 0, im.nCols() - 1, im.nRows() - 1, im.nCols() - 1); // right
        Copy(extIm, i, nPixels, im, 0, 0, 0, im.nCols() - 1); // top
        Copy(extIm, extIm.nRows() - i - 1, nPixels, im, im.nRows() - 1, 0, im.nRows() - 1, im.nCols() - 1); // bottom
    }
        
    CArray<T> colour;
    colour = PixelColour(im, 0, 0);
    DrawRect(extIm, 0, 0, nPixels - 1, nPixels - 1, colour); // top left
    colour = PixelColour(im, 0, im.nCols() - 1); 
    DrawRect(extIm, 0, extIm.nCols() - nPixels, nPixels - 1, extIm.nCols() - 1, colour); // top right
    colour = PixelColour(im, im.nRows() - 1, 0);
    DrawRect(extIm, extIm.nRows() - nPixels, 0, extIm.nRows() - 1, nPixels - 1, colour); // bottom left
    colour = PixelColour(im, im.nRows() - 1, im.nCols() - 1);
    DrawRect(extIm, extIm.nRows() - nPixels, extIm.nCols() - nPixels, extIm.nRows() - 1, extIm.nCols() - 1, colour); // bottom right

}

template <class T> void ExtendZeroOrderRB(CImage<T> &extIm, const CImage<T> &im, int nPixels)
{
    extIm.Create(im.nRows() + nPixels, im.nCols() + nPixels, im.nBands());
    
    Copy(extIm, 0, 0, im);
    
    for (int i = 0; i < nPixels; i++)
    {
        Copy(extIm, 0, extIm.nCols() - i - 1, im, 0, im.nCols() - 1, im.nRows() - 1, im.nCols() - 1); // right
        Copy(extIm, extIm.nRows() - i - 1, 0, im, im.nRows() - 1, 0, im.nRows() - 1, im.nCols() - 1); // bottom
    }
        
    CArray<T> colour;
    colour = PixelColour(im, im.nRows() - 1, im.nCols() - 1);
    DrawRect(extIm, extIm.nRows() - nPixels, extIm.nCols() - nPixels, extIm.nRows() - 1, extIm.nCols() - 1, colour); // bottom right
}

template <class T> CArray<T> PixelColour(const CImage<T> &im, int row, int col)
{
    CArray<T> colour(im.nBands());
    
    for (int b = 0; b < im.nBands(); b++)
        colour[b] = im(row, col, b);
    
    return colour;
}
    
template <class T> void DrawRect(CImage<T> &im, int row0, int col0, int row1, int col1, CArray<T> &colour)
{
    assert(im.nBands() == colour.m_nData);

    row0 = min(max(0, row0), im.nRows() - 1);
    row1 = min(max(0, row1), im.nRows() - 1);
    col0 = min(max(0, col0), im.nCols() - 1);
    col1 = min(max(0, col1), im.nCols() - 1);
    
    for (int r = row0; r <= row1; r++)
        for (int c = col0; c <= col1; c++)
            for (int b = 0; b < im.nBands(); b++)
                im(r, c, b) = colour[b];
    
}
