// MatrixBase.h

#pragma once

#include "Array.h"

//#ifndef _USE_MATH_DEFINES
//#define _USE_MATH_DEFINES
//#endif
#include <climits>
#include <cfloat>

#include "math.h"

#include <iostream>
using namespace std;

template <class T> 
class CMatrix : public CArray<T>
{
 public:
  CMatrix();
  ~CMatrix();

  CMatrix(int nRows, int nCols);
  CMatrix(int nData);

  CMatrix(const CMatrix<T> &M);
  CMatrix(const CArray<T> &A);

  CMatrix& operator=(const CMatrix<T> &M);
  CMatrix& operator=(const CArray<T> &A);
  //CMatrix& operator=(const vector<vector<T> > &m); // disabled
  CMatrix& operator=(const T t);

  void Create(int nRows, int nCols); 
  void Create(int nData);
  void Create(const CArray<int> &size);
    
  T& operator()(int row, int col);
  T& elementReference(int row, int col) const;
  const T operator()(int row, int col) const;
  const T elementCopy(int row, int col) const;

  CMatrix row(int nRow) const;
  CMatrix col(int nCol) const;

  int rowStride(void) const;
  int colStride(void) const;

  CArrayi size(void) const;
  int nRows(void) const;
  int nCols(void) const;

  void eye(void); 
};

typedef CMatrix<float>         CMatrixf;
typedef CMatrix<int>           CMatrixi;
typedef CMatrix<unsigned char> CMatrixu;
typedef CMatrix<bool>          CMatrixb;

//
// Implementation
//

template <class T>
CMatrix<T>::CMatrix()
{
    //cerr << "CMatrix<T>::CMatrix()\n";
}

template <class T>
CMatrix<T>::~CMatrix()
{
    //cerr << "CMatrix<T>::~CMatrix()\n";
}

template <class T>
CMatrix<T>::CMatrix(int nRows, int nCols)
{  
    //cerr << "CMatrix<T>::CMatrix(int nRows, int nCols)\n";
    //this->setDataNull(); // not needed as CArray<T>::CArray() is called
    Create(nRows, nCols);
}

template <class T>
CMatrix<T>::CMatrix(int nData)
{
    //cerr << "CMatrix<T>::CMatrix(int nData)\n";
    Create(nData);
}

template <class T>
CMatrix<T>::CMatrix(const CMatrix<T> &M)
{
    //cerr << "CMatrix<T>::CMatrix(const CMatrix<T> &M)\n";  
    //this->setDataNull();
    this->operator=(M);
}

template <class T>
CMatrix<T>::CMatrix(const CArray<T> &A)
{
    //this->setDataNull();
    this->operator=(A);
    //this->Make2D();
}

template <class T>
CMatrix<T>& CMatrix<T>::operator=(const CMatrix<T> &M)
{
  CArray<T> *Aptr = (CArray<T> *)(&M);
  this->operator=(*Aptr);
  this->Make2D();
  return *this;
}

template <class T>
CMatrix<T>& CMatrix<T>::operator=(const CArray<T> &A)
{
  ((CArray<T> *)this)->operator=(A);
  this->Make2D();
  return *this;
}

/*template <class T>
CMatrix<T>& CMatrix<T>::operator=(const vector<vector<T> > &m)
{
    ((CArray<T> *)this)->operator=(m);
    }*/

template <class T>
CMatrix<T>& CMatrix<T>::operator=(const T t)
{
  ((CArray<T> *)this)->operator=(t);
  this->Make2D();
  return *this;
}

template <class T>
void CMatrix<T>::Create(int nRows, int nCols)
{  
  //cerr << "CMatrix<T>::Create()" << endl;
  int nDims = 2;
  int dims[2];
  dims[0] = nRows;
  dims[1] = nCols;

  ((CArray<T> *)this)->Create(dims, nDims);
}

template <class T>
void CMatrix<T>::Create(int nData)
{
    //cerr << "CMatrix<T>::Create(int nData)\n";
    int nDims = 2;
    int dims[2];
    dims[0] = 1;
    dims[1] = nData;

    ((CArray<T> *)this)->Create(dims, nDims);
}

template <class T>
void CMatrix<T>::Create(const CArray<int> &size)
{
  Create(size[0], size[1]);
}

template <class T>
T& CMatrix<T>::operator()(int row, int col)
{
  // return elementReference(row, col);

  int index = row * this->m_stride[0] + col;
  return this->m_data[index];
}

template <class T>
T& CMatrix<T>::elementReference(int row, int col) const
{
  /*int sub[2];
  sub[0] = row;
  sub[1] = col;
  int index = this->sub2ind(sub);
  return this->m_data[index];*/

  //int index = row * m_stride[0] + col * m_stride[1];

  int index = row * this->m_stride[0] + col;
  return this->m_data[index];
}

template <class T>
const T CMatrix<T>::operator()(int row, int col) const
{
  /*int sub[2];
  sub[0] = row;
  sub[1] = col;
  int index = this->sub2ind(sub);
  return this->m_data[index];*/

  int index = row * this->m_stride[0] + col;
  return this->m_data[index];
}

template <class T>
const T CMatrix<T>::elementCopy(int row, int col) const
{
  int index = row * this->m_stride[0] + col;
  return this->m_data[index];
}

template <class T>
CMatrix<T> CMatrix<T>::col(int nCol) const
{
    const CMatrix<T> &M = *this;
    CMatrix<T> Mr;
    Mr.Create(M.nRows(), 1);
    for (int i = 0; i < M.nRows(); i++)
        Mr(i, 0) = M(i, nCol);
    return Mr;
}

template <class T>
CMatrix<T> CMatrix<T>::row(int nRow) const
{
    const CMatrix<T> &M = *this;
    CMatrix<T> Mr;
    Mr.Create(1, M.nCols());
    for (int i = 0; i < M.nCols(); i++)
        Mr(0, i) = M(nRow, i);
    return Mr;
}

template <class T>
int CMatrix<T>::rowStride(void) const
{
  return this->m_stride[0];
}

template <class T>
int CMatrix<T>::colStride(void) const
{
  return this->m_stride[1];
}

template <class T>
CArrayi CMatrix<T>::size(void) const
{
    CArrayi sz;
    
    if (this->isEmpty())
    {
        sz.Create(2);
        sz = 0;
        return sz;
    }

    return ((CArray<T>*)this)->size();
}

template <class T>
int CMatrix<T>::nRows(void) const
{
    if (this->isEmpty())
        return 0;
    
    return this->m_dims[0];
}

template <class T>
int CMatrix<T>::nCols(void) const
{
    if (this->isEmpty())
        return 0;

    return this->m_dims[1];
}

template <class T>
void CMatrix<T>::eye(void)
{
  CMatrix<T>& M = *this;
  M.zeros();
  int n = min(nRows(), nCols());
  for (int i = 0; i < n; i++)
    {
      M(i, i) = 1;
    }
}

