// MatrixFun.h

#pragma once

#include <assert.h>

// unary matrix operations

template <class T> void Transpose(CMatrix<T> &M, const CMatrix<T> &M1); // transpose
template <class T> CMatrix<T> Transpose(const CMatrix<T> &M1);
template <class T> CMatrix<T> operator~(const CMatrix<T> &M1);

// binary matrix operations

template <class T> void Multiply(CMatrix<T> &M, const CMatrix<T> &M1, const CMatrix<T> &M2); // matrix multiply
template <class T> CMatrix<T> operator*(const CMatrix<T> &M1, const CMatrix<T> &M2);

// matrix manipulation

template <class T> CMatrix<T> RowSum(const CMatrix<T> &M1);
template <class T> CMatrix<T> ColSum(const CMatrix<T> &M1);
template <class T> CMatrix<T> RowMean(const CMatrix<T> &M1);
template <class T> CMatrix<T> ColMean(const CMatrix<T> &M1);
template <class T> T          Dot(const CMatrix<T> &M1, const CMatrix<T> &M2);
template <class T> CMatrix<T> DotCols(const CMatrix<T> &M1, const CMatrix<T> &M2);
template <class T> CMatrix<T> Cross(const CMatrix<T> &a, const CMatrix<T> &b);

template <class T> CMatrix<T> Diag(const CArray<T> &d);
template <class T> void Diag(CMatrix<T> &M, const CArray<T> &d);
template <class T> CArray<T> Diag(const CMatrix<T> &M);
template <class T> void Diag(CArray<T> &d, const CMatrix<T> &M);

// convolve, copy and select

//template <class T> void Convolve(T* dstptr, T* srcptr, int nData, const CArray<T> &kernel); 
template <class T> void ConvolveRows(CMatrix<T> &M, const CMatrix<T> &M1, const CArray<T> &kernel);
template <class T> void ConvolveRows(CMatrix<T> &M, const CArray<T> &kernel); // in place
template <class T> void ConvolveCols(CMatrix<T> &M, const CMatrix<T> &M1, const CArray<T> &kernel);
template <class T> void ConvolveCols(CMatrix<T> &M, const CArray<T> &kernel); 
template <class T> void ConvolveSeparable(CMatrix<T> &M, const CMatrix<T> &M1, const CArray<T> &kernel);
template <class T> void ConvolveSeparable(CMatrix<T> &M, const CArray<T> &kernel);

template <class T> void IntegralRows(CMatrix<T> &M); // in place
template <class T> void IntegralRows(CMatrix<T> &Mdst, const CMatrix<T> &Msrc);
template <class T> void IntegralCols(CMatrix<T> &M); 
template <class T> void IntegralCols(CMatrix<T> &Mdst, const CMatrix<T> &Msrc);
template <class T> void Integral(CMatrix<T> &M);
template <class T> void Integral(CMatrix<T> &Mdst, const CMatrix<T> &Msrc);

template <class T> CArrayi FirstNonZeroRows(const CMatrix<T> &M);
template <class T> CArrayi LastNonZeroRows(const CMatrix<T> &M);
template <class T> CArrayi FirstNonZeroCols(const CMatrix<T> &M);
template <class T> CArrayi LastNonZeroCols(const CMatrix<T> &M);

template <class T> void Copy(CMatrix<T> &MDst, int rowDst, int colDst, 
			     const CMatrix<T> &MSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2);
template <class T> void Copy(CMatrix<T> &MDst, int rowDst, int colDst, const CMatrix<T> &MSrc);
//template <class T> void Copy(CMatrix<T> &MDst, const CMatrix<T> &MSrc);

template <class T> void Add(CMatrix<T> &MDst, int rowDst, int colDst, 
			     const CMatrix<T> &MSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2);
template <class T> void Add(CMatrix<T> &MDst, int rowDst, int colDst, const CMatrix<T> &MSrc);

template <class T> void Select(CMatrix<T> &MDst, const CMatrix<T> &MSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2);
template <class T> CMatrix<T> Select(const CMatrix<T> &MSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2);

template <class T> void SelectRows(CMatrix<T> &MDst, const CMatrix<T> &MSrc, CArrayi rows);
template <class T> CMatrix<T> SelectRows(const CMatrix<T> &MSrc, CArrayi rows);
template <class T> void SelectCols(CMatrix<T> &MDst, const CMatrix<T> &MSrc, CArrayi rows);
template <class T> CMatrix<T> SelectCols(const CMatrix<T> &MSrc, CArrayi rows);

template <class T> void SelectRow(CMatrix<T> &MDst, const CMatrix<T> &MSrc, int row);
template <class T> CMatrix<T> SelectRow(const CMatrix<T> &MSrc, int row);
template <class T> void SelectCol(CMatrix<T> &MDst, const CMatrix<T> &MSrc, int col);
template <class T> CMatrix<T> SelectCol(const CMatrix<T> &MSrc, int col);

template <class T> void SelectRow1(CArray<T> &MDst, const CMatrix<T> &MSrc, int row);
template <class T> CArray<T> SelectRow1(const CMatrix<T> &MSrc, int row);
template <class T> void SelectCol1(CArray<T> &MDst, const CMatrix<T> &MSrc, int col);
template <class T> CArray<T> SelectCol1(const CMatrix<T> &MSrc, int col);

template <class T> void Repmat(CMatrix<T> &M, const CMatrix<T> &M1, int rows, int cols);
template <class T> CMatrix<T> Repmat(const CMatrix<T> &M1, int rows, int cols);

// convert to std vector

template <class T> void Convert(vector<vector<T> > &Mv, const CMatrix<T> &M); // note inner vectors are rows of M
template <class T> void Convert(CMatrix<T> &M, const vector<vector<T> > &m);

//
// Implementation
//

// unary matrix operations

template <class T>
void Transpose(CMatrix<T> &M, const CMatrix<T> &M1)
{
    M.Create(M1.nCols(), M1.nRows());

    for (int r = 0; r < M.nRows(); r++)
    {
        for (int c = 0; c < M.nCols(); c++)
	{
            M(r, c) = M1(c, r);
	}
    }
}

template <class T>
CMatrix<T> Transpose(const CMatrix<T> &M1)
{
    CMatrix<T> M;
    Transpose(M, M1);
    return M;   
}

template <class T>
CMatrix<T> operator~(const CMatrix<T> &M1)
{
    CMatrix<T> M;
    Transpose(M, M1);
    return M;   
}

// binary matrix operations

template <class T>
void Multiply(CMatrix<T> &M, const CMatrix<T> &M1, const CMatrix<T> &M2)
{
    assert(M1.nCols() == M2.nRows());
  
    int n = M1.nCols();

    M.Create(M1.nRows(), M2.nCols());
    
    for (int i = 0; i < M1.nRows(); i++)
    {
        for (int j = 0; j < M2.nCols(); j++)
	{
            for (int k = 0; k < n; k++)
	    {
                M(i, j) += M1(i, k) * M2(k, j);
	    }
	}
    }
}


template <class T>
CMatrix<T> operator*(const CMatrix<T> &M1, const CMatrix<T> &M2)
{
    CMatrix<T> M;
    Multiply(M, M1, M2);
    return M;
}


// matrix manipulation

template <class T>
CMatrix<T> RowSum(const CMatrix<T> &M1)
{
    CMatrix<T> M(1, M1.nCols());

    for (int j = 0; j < M1.nCols(); j++)
    {
        float rowsumj = 0;
        for (int i = 0; i < M1.nRows(); i++)
	{
            rowsumj += M1(i, j);
	}
        M(0, j) = rowsumj;
    }

    return(M);
}

template <class T>
CMatrix<T> ColSum(const CMatrix<T> &M1)
{
    CMatrix<T> M(M1.nRows(), 1);

    for (int i = 0; i < M1.nRows(); i++)
    {
        float colsumi = 0;
        for (int j = 0; j < M1.nCols(); j++)
	{
            colsumi += M1(i, j);
	}
        M(i, 0) = colsumi;
    }

    return(M);
}

template <class T> CMatrix<T> RowMean(const CMatrix<T> &M1)
{
    return (RowSum(M1)/(float)M1.nRows());
}

template <class T> CMatrix<T> ColMean(const CMatrix<T> &M1)
{
    return (ColSum(M1)/(float)M1.nCols());
}


template <class T>
T Dot(const CMatrix<T> &M1, const CMatrix<T> &M2)
{
    assert(M1.nRows() == M2.nRows());
    assert(M1.nCols() == 1);
    assert(M2.nCols() == 1);

    return(Sum(ScalarMult(M1, M2)));
}

template <class T>
CMatrix<T> DotCols(const CMatrix<T> &M1, const CMatrix<T> &M2)
{
    assert(M1.nRows() == M2.nRows());
    assert(M1.nCols() == M2.nCols());

    CMatrix<T> M12;
    ScalarMult(M12, M1, M2);
    return (RowSum(M12));
    
    //return(RowSum(ScalarMult(M1, M2)));
}

template <class T>
CMatrix<T> Cross(const CMatrix<T> &a, const CMatrix<T> &b)
{
    assert(a.nRows() == 3);
    assert(b.nRows() == 3);
    assert(a.nCols() == 1);
    assert(b.nCols() == 1);

    CMatrix<T> c(3, 1);
    c[0] = a[1] * b[2] - b[1] * a[2];
    c[1] = a[2] * b[0] - b[2] * a[0];
    c[2] = a[0] * b[1] - b[0] * a[1];

    return(c);
}

template <class T> void Diag(CMatrix<T> &M, const CArray<T> &d)
{
    int n = d.nData();
    M.Create(n, n);
    for (int i = 0; i < n; i++)
        M(i, i) = d[i];
}

template <class T> CMatrix<T> Diag(const CArray<T> &d)
{
    CMatrix<T> M;
    Diag(M, d);
    return M;
}

template <class T> void Diag(CArray<T> &d, const CMatrix<T> &M)
{
    int n = min(M.nRows(), M.nCols());
    d.Create(n);
    for (int i = 0; i < n; i++)
        d[i] = M(i, i);
}

template <class T> CArray<T> Diag(const CMatrix<T> &M)
{
    CArray<T> d;
    Diag(d, M);
    return d;
}

// convolve, copy and select

/*template <class T> void Convolve(T* dstptr, T* srcptr, int nData, const CArray<T> &kernel)
  {      
  int ksize = kernel.m_nData;
  T* kptr0  = kernel.m_data;
    
  // make sure kernel length is odd
  CArray<T> kernel2;
  if (ksize % 2 == 0)
  {
  kernel2.Create(ksize + 1);
  for (int i = 0; i < kernel.m_nData; i++)
  kernel2[i] = kernel[i];
  kptr0 = kernel2.m_data;
  ksize++;
  }
  
  int length = (ksize - 1)/2;
  int nDataBuffer = nData + 2 * length;

  // extend data outside range
  CArray<T> buffer(nDataBuffer);
  T *bufferptr = buffer.m_data;
  
  // extend left
  for (int i = 0; i < length; i++)
  {
  memcpy(bufferptr, srcptr, sizeof(T));      
  bufferptr++;
  }
	
  // copy centre
  memcpy(bufferptr, srcptr, sizeof(T));
  bufferptr += nData;
  srcptr += nData;

  // extend right
  for (int i = 0; i < length; i++)
  {
  memcpy(bufferptr, srcptr, sizeof(T));
  bufferptr++;
  }

  memset(dstptr, 0, nData * sizeof(T)); // needed for in-place use 
       
  // Convolve buffer with kernel    
  T *kptr;
  T *kptr1      = kptr0 + ksize - 1;
  T *bufferptr0 = buffer.m_data;

  for (int i = 0; i < nData; i++)
  {
  kptr      = kptr1;
  bufferptr = bufferptr0;
      
  for (int k = 0; k < ksize; k++)
  {  	  
  *dstptr += *bufferptr * *kptr;
  bufferptr++;
  kptr--;
  }

  dstptr++;
  bufferptr0++;
  }
  
  }*/

template <class T> void ConvolveRows(CMatrix<T> &M, const CMatrix<T> &M1, const CArray<T> &kernel)
{
    M.Create(M1.size());

    T* srcptr = M1.m_data;
    T* dstptr = M.m_data;

    for (int r = 0; r < M.nRows(); r++)
    {
        Convolve(dstptr, srcptr, M.nCols(), kernel);      
        srcptr += M1.rowStride();
        dstptr += M.rowStride();
    }

}

template <class T> void ConvolveRows(CMatrix<T> &M, const CArray<T> &kernel)
{  
    T* rowptr = M.m_data;

    for (int r = 0; r < M.nRows(); r++)
    {
        Convolve(rowptr, rowptr, M.nCols(), kernel);
        rowptr += M.rowStride();
    }
}

template <class T> void ConvolveCols(CMatrix<T> &M, const CMatrix<T> &M1, const CArray<T> &kernel)
{
    CMatrix<T> Mt;
    Transpose(Mt, M1);
    ConvolveRows(Mt, kernel);  
    Transpose(M, Mt);
}

template <class T> void ConvolveCols(CMatrix<T> &M, const CArray<T> &kernel)
{
    CMatrix<T> Mt;
    Transpose(Mt, M);
    ConvolveRows(Mt, kernel);
    Transpose(M, Mt);
}

template <class T> void ConvolveSeparable(CMatrix<T> &M, const CMatrix<T> &M1, const CArray<T> &kernel)
{  
    ConvolveRows(M, M1, kernel);
    ConvolveCols(M, kernel);
}

template <class T> void ConvolveSeparable(CMatrix<T> &M, const CArray<T> &kernel)
{  
    ConvolveRows(M, kernel);
    ConvolveCols(M, kernel);
}

template <class T> 
void IntegralRows(CMatrix<T> &M)
{
  for (int r = 0; r < M.nRows(); r++)
    {
      for (int c = 0; c < (M.nCols() - 1); c++)
	{
	  M(r, c + 1) += M(r, c);
	}
    }
}

template <class T> 
void IntegralRows(CMatrix<T> &Mdst, const CMatrix<T> &Msrc)
{
  Mdst = Msrc;
  IntegralRows(Mdst);
}

/*template <class T> CMatrix<T> IntegralRows(const CMatrix<T> &M)
{
  CMatrix<T> Mr;
  IntegralRows(Mr, M);
  return Mr;
  }*/

template <class T> 
void IntegralCols(CMatrix<T> &M)
{
  for (int c = 0; c < M.nCols(); c++)
    {
      for (int r = 0; r < (M.nRows() - 1); r++)
	{
	  M(r + 1, c) += M(r, c);
	}
    }
}

template <class T>
void IntegralCols(CMatrix<T> &Mdst, const CMatrix<T> &Msrc)
{
  Mdst = Msrc;
  IntegralCols(Mdst);
}

/*template <class T> CMatrix<T> IntegralCols(const CMatrix<T> &M)
{
  CMatrix<T> Mr;
  IntegralCols(Mr, M);
  return Mr;
  }*/

template <class T> 
void Integral(CMatrix<T> &M)
{
  IntegralRows(M);
  IntegralCols(M);
}

template <class T>
void Integral(CMatrix<T> &Mdst, const CMatrix<T> &Msrc)
{
  IntegralRows(Mdst, Msrc);
  IntegralCols(Mdst);
}

template <class T>
CArrayi FirstNonZeroRows(const CMatrix<T> &M)
{
    CArrayi idx(M.nRows());

    for (int r = 0; r < M.nRows(); r++)
    {
        idx[r] = -1;
        bool done = false;
        for (int c = 0; c < M.nCols() && !done; c++)
        {
            if (M(r, c) != 0)
            {
                idx[r] = c;
                done = true;
            }
        }
    }
    
    return idx;
}

template <class T>
CArrayi LastNonZeroRows(const CMatrix<T> &M)
{
    CArrayi idx(M.nRows());

    for (int r = 0; r < M.nRows(); r++)
    {
        idx[r] = -1;
        bool done = false;
        for (int c = M.nCols() - 1; c >= 0 && !done; c--)
        {
            if (M(r, c) != 0)
            {
                idx[r] = c;
                done = true;
            }
        }
    }
    
    return idx;
}

template <class T>
CArrayi FirstNonZeroCols(const CMatrix<T> &M)
{
    CArrayi idx(M.nCols());

    for (int c = 0; c < M.nCols(); c++)
    {
        idx[c] = -1;
        bool done = false;
        for (int r = 0; r < M.nRows() && !done; r++)
        {
            if (M(r, c) != 0)
            {
                idx[c] = r;
                done = true;
            }
        }
    }
    
    return idx;
}

template <class T>
CArrayi LastNonZeroCols(const CMatrix<T> &M)
{
    CArrayi idx(M.nCols());
    
    for (int c = 0; c < M.nCols(); c++)
    {
        idx[c] = -1;
        bool done = false;
        for (int r = M.nRows() - 1; r >= 0 && !done; r--)
        {
            if (M(r, c) != 0)
            {
                idx[c] = r;
                done = true;
            }
        }
    }
    
    return idx;
}

template <class T> void Copy(CMatrix<T> &MDst, int rowDst, int colDst, 
                             const CMatrix<T> &MSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2)
{
  
    int nRowsCopy  = min(MDst.nRows() - rowDst, rowSrc2 - rowSrc1 + 1);
    int nColsCopy  = min(MDst.nCols() - colDst, colSrc2 - colSrc1 + 1);
    
    rowSrc2 = rowSrc1 + nRowsCopy - 1;
    colSrc2 = colSrc1 + nColsCopy - 1;
    
    T* srcRowPtr = &MSrc.elementReference(rowSrc1, colSrc1);
    T* dstRowPtr = &MDst(rowDst, colDst);

    int nBytesPerRow = nColsCopy * sizeof(T);

    for (int r = 0; r < nRowsCopy; r++)
    {
        memcpy(dstRowPtr, srcRowPtr, nBytesPerRow);
        srcRowPtr += MSrc.rowStride();
        dstRowPtr += MDst.rowStride();
    }
}

template <class T> void Copy(CMatrix<T> &MDst, int rowDst, int colDst, const CMatrix<T> &MSrc)
{
    Copy(MDst, rowDst, colDst, MSrc, 0, 0, MSrc.nRows() - 1, MSrc.nCols() - 1);
}

/*template <class T> void Copy(CMatrix<T> &MDst, const CMatrix<T> &MSrc)
{
    Copy(MDst, 0, 0, MSrc, 0, 0, MSrc.nRows() - 1, MSrc.nCols() - 1);
    }*/

template <class T> void Add(CMatrix<T> &MDst, int rowDst, int colDst, 
                             const CMatrix<T> &MSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2)
{  
    int nRowsAdd  = min(MDst.nRows() - rowDst, rowSrc2 - rowSrc1 + 1);
    int nColsAdd  = min(MDst.nCols() - colDst, colSrc2 - colSrc1 + 1);
    
    rowSrc2 = rowSrc1 + nRowsAdd - 1;
    colSrc2 = colSrc1 + nColsAdd - 1;
    
    T* srcRowPtr = &MSrc.elementReference(rowSrc1, colSrc1);
    T* dstRowPtr = &MDst(rowDst, colDst);

    for (int r = 0; r < nRowsAdd; r++)
    {
        T* srcPtr = srcRowPtr;
        T* dstPtr = dstRowPtr;
        for (int c = 0; c < nColsAdd; c++)
        {
            *dstPtr += *srcPtr;
            dstPtr++;
            srcPtr++;
        }
        srcRowPtr += MSrc.rowStride();
        dstRowPtr += MDst.rowStride();
    }
}

template <class T> void Add(CMatrix<T> &MDst, int rowDst, int colDst, const CMatrix<T> &MSrc)
{
    Add(MDst, rowDst, colDst, MSrc, 0, 0, MSrc.nRows() - 1, MSrc.nCols() - 1);
}

template <class T> void Select(CMatrix<T> &MDst, const CMatrix<T> &MSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2)
{

    int nRowsCopy = rowSrc2 - rowSrc1 + 1;
    int nColsCopy = colSrc2 - colSrc1 + 1;

    MDst.Create(nRowsCopy, nColsCopy);
    
    T* srcRowPtr = &MSrc.elementReference(rowSrc1, colSrc1);
    T* dstRowPtr = MDst.m_data;

    int nBytesPerRow = nColsCopy * sizeof(T);

    for (int r = 0; r < nRowsCopy; r++)
    {
        memcpy(dstRowPtr, srcRowPtr, nBytesPerRow);
        srcRowPtr += MSrc.rowStride();
        dstRowPtr += MDst.rowStride();
    }

}

template <class T> CMatrix<T> Select(const CMatrix<T> &MSrc, int rowSrc1, int colSrc1, int rowSrc2, int colSrc2)
{
    CMatrix<T> MDst;
    Select(MDst, MSrc, rowSrc1, colSrc1, rowSrc2, colSrc2);
    return MDst;
}

template <class T> void SelectRows(CMatrix<T> &MDst, const CMatrix<T> &MSrc, CArrayi rows)
{
    int nRows = rows.m_nData;    
    MDst.Create(nRows, MSrc.nCols());
    
    for (int i = 0; i < nRows; i++)
        Copy(MDst, i, 0, MSrc, rows[i], 0, rows[i], MSrc.nCols() - 1);    
}

template <class T> CMatrix<T> SelectRows(const CMatrix<T> &MSrc, CArrayi rows)
{
    CMatrix<T> MDst; 
    SelectRows(MDst, MSrc, rows);
    return MDst;
}

template <class T> void SelectCols(CMatrix<T> &MDst, const CMatrix<T> &MSrc, CArrayi cols)
{
    int nCols = cols.m_nData;    
    MDst.Create(MSrc.nRows(), nCols);
    
    for (int i = 0; i < nCols; i++)
        Copy(MDst, 0, i, MSrc, 0, cols[i], MSrc.nRows() - 1, cols[i]);    
}

template <class T> CMatrix<T> SelectCols(const CMatrix<T> &MSrc, CArrayi cols)
{
    CMatrix<T> MDst; 
    SelectCols(MDst, MSrc, cols);
    return MDst;
}

template <class T> void SelectRow(CMatrix<T> &MDst, const CMatrix<T> &MSrc, int row)
{
    Select(MDst, MSrc, row, 0, row, MSrc.nCols() - 1);
}

template <class T> CMatrix<T> SelectRow(const CMatrix<T> &MSrc, int row)
{
    CMatrix<T> MDst;
    SelectRow(MDst, MSrc, row);
    return MDst;
}

template <class T> void SelectCol(CMatrix<T> &MDst, const CMatrix<T> &MSrc, int col)
{
    Select(MDst, MSrc, 0, col, MSrc.nRows() - 1, col);
}

template <class T> CMatrix<T> SelectCol(const CMatrix<T> &MSrc, int col)
{
    CMatrix<T> MDst;    
    SelectCol(MDst, MSrc, col);
    return MDst;
}


template <class T> void SelectRow1(CArray<T> &ADst, const CMatrix<T> &MSrc, int row)
{
    ADst.Create(MSrc.nCols());
    memcpy(ADst.m_data, &MSrc.elementReference(row, 0), MSrc.nCols() * sizeof(T)); 
}

template <class T> CArray<T> SelectRow1(const CMatrix<T> &MSrc, int row)
{
    CArray<T> ADst;
    SelectRow(ADst, MSrc, row);
    return ADst;
}

template <class T> void SelectCol1(CArray<T> &ADst, const CMatrix<T> &MSrc, int col)
{
    ADst.Create(MSrc.nRows());
    for (int i = 0; i < MSrc.nRows(); i++)
        ADst[i] = MSrc(i, col);
}

template <class T> CArray<T> SelectCol1(const CMatrix<T> &MSrc, int col)
{
    CArray<T> ADst;    
    SelectCol(ADst, MSrc, col);
    return ADst;
}

template <class T>
void Repmat(CMatrix<T> &M, const CMatrix<T> &M1, int rows, int cols)
{
    int nRowsr = M1.nRows() * rows;
    int nColsr = M1.nCols() * cols;
    M.Create(nRowsr, nColsr);

    for (int i = 0; i < rows; i++)
    {
        for (int j = 0; j < cols; j++)
        {
            Copy(M, i * M1.nRows(), j * M1.nCols(), M1);
        }
    }

}

template <class T>
CMatrix<T> Repmat(const CMatrix<T> &M1, int rows, int cols)
{
    CMatrix<T> M;
    Repmat(M, M1, rows, cols);
    return M;
}

// convert to/from std vector

template <class T> 
void Convert(vector<vector<T> > &Mv, const CMatrix<T> &M)
{
    Mv.resize(M.nRows());
    for (int r = 0; r < M.nRows(); r++)
        Mv[r].resize(M.nCols());
    
    for (int r = 0; r < M.nRows(); r++)
        for (int c = 0; c < M.nCols(); c++)
            Mv[r][c] = M(r, c);
}

template <class T>
void Convert(CMatrix<T> &M, const vector<vector<T> > &m)
{
  int nRows = (int)m.size();
  if (nRows == 0)
    return;

  int nCols = (int)m[0].size();
  if (nCols == 0)
    return;

  bool isValid = true;
  for (int i = 0; i < nRows; i++)
    if (m[i].size() != nCols)
      isValid = false;

  if (!isValid)
    return;

  /*int nDims = 2;
  int dims[2];
  dims[0] = nRows;
  dims[1] = nCols;
  
  A.Create(dims, nDims);*/

  M.Create(nRows, nCols);

  //  int n = 0;
  for (int r = 0; r < nRows; r++)
    for (int c = 0; c < nCols; c++)
      {
	M(r, c) = m[r][c];
	//M.m_data[n] = m[r][c];
	//n++;
      }
}
