// main.cpp

// imalign im1 im1Filename im2 im2Filename

#include "SIFTMatch.h"
#include "ImageFile.h"
#include "helpers.h"
#include <iostream>
#include <fstream>

using namespace std;

int main(int argc, char* argv[])
{	  	
    // Set default parameters
    CParameter Param;
    Param.Set("im1", "0.jpg");
    Param.Set("im2", "1.jpg");    
    Param.Set("out1", "./imalign/red.tif");
    Param.Set("out2", "./imalign/nir.tif");

    // SIFT param
    Param.Set("minImageDim", 1000);
    Param.Set("skipLevels", 2);
    Param.Set("nOverlap", 2);
    Param.Set("outlierDist", 0.5f);
    Param.Set("numBB", 200);
    Param.Set("minNInliers", 40);

    // RANSAC param
    Param.Set("ransac_nIterations", 1500);
    Param.Set("ransac_epsilon", 0.02f);
    
    Param.Set("drawInterestPoints", false);

    // Get command line params
    CParameter ParamCmd;
    ParamCmd.ReadCommandLine(argc, argv);    
    Param.Set(ParamCmd);
 
//    int dum;
//    dum = system("mkdir results 1>/dev/null 2>/dev/null");
   
    // Match SIFT features
    cerr << "[ Match SIFT ]\n";
    vector<CImageFilef> im;
    vector<CMatrixf> ip;
    CArrayf ipExtractScale;
    vector<vector<CMatrixi> > matches;
    CMatrixi nMatches;

    im.resize(2);
    string im1Filename = Param.Gets("im1");
    string im2Filename = Param.Gets("im2");
    im[0].Create(im1Filename, "im0");
    im[1].Create(im2Filename, "im1");
        
    //SIFTMatch(im, ip, matches, nMatches, Param);
    SIFTMatch(im, ip, ipExtractScale, matches, nMatches, Param);

    // RANSAC for similarity transform
    cerr << "[ Perform RANSAC ]\n";

    const CMatrixi &match12 = matches[0][1];
    int nMatches12 = match12.nCols();

    CMatrixf u1(2, nMatches12);
    CMatrixf u2(2, nMatches12);

    for (int i = 0; i < nMatches12; i++)
    {
        int ID1 = match12(0, i);
        int ID2 = match12(1, i);
        
        u1(0, i) = ip[0](0, ID1);
        u1(1, i) = ip[0](1, ID1);
        
        u2(0, i) = ip[1](0, ID2);
        u2(1, i) = ip[1](1, ID2);
    }

    CArrayi inliersID;
    CMatrixf S;
    
    CParameter RANSACParam;
    RANSACParam.SetNamed(Param, "ransac");
    CArrayi imSize1 = im[0].OriginalSize();
    CArrayi imSize2 = im[1].OriginalSize();

    RANSACSimilarity(inliersID, S, u1, u2, imSize1, imSize2, RANSACParam);    

    CMatrixf u1i = SelectCols(u1, inliersID);
    CMatrixf u2i = SelectCols(u2, inliersID);

    // Recompute S from inliers
    S = ComputeSimilarity(u1i, u2i);
    
/*    if (Param.Getb("drawInterestPoints"))
    {
        CImagef im1t = im[0];
        CImagef im2t = im[1];

        // Draw interest points
        DrawInterestPoints(im1t, ip1i);
        DrawInterestPoints(im2t, ip2i);
        
        // Write images
        WriteJPEG(im1t, "results/ransac1.jpg");
        WriteJPEG(im2t, "results/ransac2.jpg");
        }   */

    int nInliers = u1i.nCols();
    printf(" Found %d inliers\n", nInliers);
    //cerr << " S = \n" << S.Dump();

    // Require min num of inliers
    if (nInliers < Param.Geti("minNInliers"))
    {
        printf(" *** rejecting match *** \n");
        return (0);
    }
    
    // Resample image 2 in image 1 coords
    //im[0].Get();
    //im[1].Get();
    CImagef im0, im1;
    im[0].Read(im0);
    im[1].Read(im1);

    cerr << "[ Resample image ]\n";
    CImagef imCoord;    
    ComputeImageCoords(imCoord, im[0].OriginalSize(), S);

    CImagef im2r;
    CImagei alphaIm;
    SampleImageBilinear(im2r, alphaIm, imCoord, im1);

    // Find crop rectangle
    cerr << "[ Find Crop Rectangle ]\n";
    
    int maxSz = 512;
    CImagei alphaImt;
    int scale = 1;
    while (alphaIm.nRows() >= 2 * maxSz || alphaIm.nCols() >= 2 * maxSz)
    {
        DownSample2_3x3(alphaImt, alphaIm);
        Swap(alphaImt, alphaIm);
        scale *= 2;
    }
    
    CArrayi rect = FindLargestRect(alphaIm);
    rect *= scale;
    cerr << " Crop rectangle = " << rect.Dump();

//Write the coordinates to a txt-file that can later be recovered in Matlab
ofstream rect_file;
rect_file.open("./imalign/rect_coords.txt");
rect_file << rect.Dump();
rect_file.close();
    
        
    // Draw crop rect on image
    // CArrayf rectf;
    // Convert(rectf, rect);
    // DrawRect2(im2r, rectf[0], rectf[1], rectf[2], rectf[3]);
    // WriteJPEG(im2r, "results/im2r.jpg");

    CImagef im1c, im2c;
    
    Select(im1c, im0, rect[0], rect[1], rect[2], rect[3]);
    Select(im2c, im2r, rect[0], rect[1], rect[2], rect[3]);

    WriteImage(im1c, Param.Gets("out1"));
    WriteImage(im2c, Param.Gets("out2"));
    
    //WriteTIFF(im1c, "results/im1.tif");
    //WriteTIFF(im2c, "results/im2.tif");

    return(1);
}
