%*************************************************************************%
% Author: Dominic Rfenacht, April 2012
% Purpose:
% Main function for the camera processing pipeline. 
%*************************************************************************%
%Input: - RAW images (visible and/or NIR)
%Output: - im_out: Developed images (visible and/or NIR)
%*************************************************************************%

function varargout = workflow_gui(varargin)
% Begin initialization code - DO NOT EDIT
gui_Singleton = 1;
gui_State = struct('gui_Name',       mfilename, ...
    'gui_Singleton',  gui_Singleton, ...
    'gui_OpeningFcn', @workflow_gui_OpeningFcn, ...
    'gui_OutputFcn',  @workflow_gui_OutputFcn, ...
    'gui_LayoutFcn',  [] , ...
    'gui_Callback',   []);
if nargin && ischar(varargin{1})
    gui_State.gui_Callback = str2func(varargin{1});
end

if nargout
    [varargout{1:nargout}] = gui_mainfcn(gui_State, varargin{:});
else
    gui_mainfcn(gui_State, varargin{:});
end
% End initialization code - DO NOT EDIT


% --- Executes just before workflow_gui is made visible. Load all needed
% variables, and prepare the selected items for initial display.
function workflow_gui_OpeningFcn(hObject, eventdata, handles, varargin)
% Choose default command line output for workflow_gui
handles.output = hObject;

load cameras.mat
%Store the cell containing the camera models along with their 3x3 cRGB2XYZ
%transformation matrices
handles.cameras = mm;
clear mm;

set(handles.popupmenu_cameraSelectionVIS, 'String', handles.cameras(:,1));
set(handles.popupmenu_cameraSelectionNIR, 'String', handles.cameras(:,1));

%In the beginning, the first camera model is selected
handles.cameramodelCam1 = handles.cameras(1,1);
handles.cameramodelCam2 = handles.cameras(1,1);
set(handles.radiobutton_all, 'Value', 1);
handles.nir_info = 'all';

set(handles.radiobutton_twoRAW, 'Value', 1);
handles.input_type = 'tworaw';

set(handles.radiobutton_WBfirst, 'Value', 1);
handles.order = 'W-D';

set(handles.radiobutton_PNG, 'Value', 1);
handles.output_format = 'png';

%Specify the White Balancing modes that are implemented
wb_modes = {'Grey World', 'Modified Grey World', 'MaxRGB', 'None'};
handles.wbMode = wb_modes{1};
set(handles.popupmenu_wbmode, 'String', wb_modes);

%Specify the demosaicing methods that are implemented
demosaicing_methods = {'Amaze', 'Bilinear interpolation', 'Subsample by 2', 'Subsample by 4', 'Subsample by 8'};
handles.demosaicing_method = demosaicing_methods{1};
set(handles.popupmenu_demosaicingMethod, 'String', demosaicing_methods);

%Specify the tone mapping methods that are implemented
tonemapping_methods = {'Linear', 'Local (Retinex)', 'Histogram equalization', 'Colorimetric reproduction'};
handles.tonemapping_method = tonemapping_methods{1};
set(handles.popupmenu_tonemapping, 'String', tonemapping_methods);

set(handles.button_startProcessing, 'Visible', 'off');

set(handles.radiobutton_VIS, 'Visible', 'off');
set(handles.radiobutton_NIR, 'Visible', 'off');
set(handles.radiobutton_VISNIR, 'Visible', 'off');
set(handles.text_cfatype, 'Visible', 'off');

set(handles.radiobutton_VIS, 'Value', 1);
set(handles.radiobutton_NIR, 'Value', 0);
set(handles.radiobutton_VISNIR, 'Value', 0);
handles.singletype = 'vis';

set(handles.radiobutton_gammaFixed, 'Value', 1);
set(handles.pushbutton_setGamma, 'Visible', 'off');
handles.gamma = 'fixed';
handles.gamma_value = 1.0;
handles.continue_workflow = 1;

%By default, don't rotate and don't align images
handles.align_images = 0;
handles.rotate_image = 0;

set(handles.checkbox_showResults, 'Value', 1);
handles.show_images = 1;

%Slider
set(handles.slider_gamma, 'Value', 1.0);

if(ismac)
    set(handles.checkbox_alignImages, 'Enable', 'on');
    handles.ismac = 1;
else
    set(handles.checkbox_alignImages, 'Enable', 'off');
    handles.ismac = 0;
end

%Default quantiles for B/W correction
handles.lower_quantile = 0.025;
set(handles.edit_quantileLower, 'String', handles.lower_quantile);
handles.upper_quantile = 0.975;
set(handles.edit_quantileUpper, 'String', handles.upper_quantile);

%This is used to remove the ugly axes that are displayed before the first
%image is shown
axes(handles.axes_image);
imshow(0.94);
axes(handles.axes_nir);
imshow(0.94);

%Default output directory
set(handles.edit_OutputDir, 'String', './Processed/')

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in button_loadRAWImage.
function button_loadRAWImage_Callback(hObject, eventdata, handles)
set(handles.button_startProcessing, 'Visible', 'off');
drawnow

%Lauch a pop-up dialog which allows selecting multiple images
[handles.filename, handles.pathname] = uigetfile( ...
    {'*.CR2;*.cr2;*.NEF;*.nef;*.CRW','RAW file (*.CR2, *.CRW, *.NEF)'; ...
    '*.CR2;*.cr2','Canon RAW (*.CR2)'; ...
    '*.NEF;*.nef',  'Nikon RAW (*.NEF)'; ...
    '*.*',  'All (*.*)'}, ...
    'Please select a RAW File', 'MultiSelect', 'on');

%If only one image is selected, uigetfile returns "String" instead of
%cells. To unify the processing, we convert this string to a cell
handles.filename = cellstr(handles.filename);

if(strcmp(handles.input_type, 'tworaw'))
    handles.pathname2 = uigetdir('./', 'Select the folder containing the NIR images');
    handles.pathname2 = [handles.pathname2 '/'];
end

if(length(handles.filename) < 2)
    updateInfo(handles, 'You selected 1 image to be processed...');
else
    updateInfo(handles, ['You selected ' num2str(length(handles.filename)) ' images to be processed...']);
end

%Display the start processing button
set(handles.button_startProcessing, 'Visible', 'on');
set(handles.button_startProcessing, 'Enable', 'on');

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in button_startProcessing.
function button_startProcessing_Callback(hObject, eventdata, handles)
set(handles.button_startProcessing, 'Enable', 'off');

%Load camera related parameters
if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'nir'))
    %Get the row in the cameras matrix corresponding to the selected entry
    handles.rowCam1 = find(strcmp(handles.cameras(:,1), handles.cameramodelCam1) == 1);
end

if(strcmp(handles.input_type, 'tworaw'))
    %Get the row in the cameras matrix corresponding to the selected entry
    handles.rowCam1 = find(strcmp(handles.cameras(:,1), handles.cameramodelCam1) == 1);   
    handles.rowCam2 = find(strcmp(handles.cameras(:,1), handles.cameramodelCam2) == 1);
end

handles.cRGB2sRGBmat = handles.cameras{handles.rowCam1, 2};
handles.cfa_pattern = handles.cameras{handles.rowCam1, 3};


%Compute the cfa indices based on the cfa pattern
cfa_ind = get_cfa_index(handles.cfa_pattern);

%Compute the pre-multipliers based on the quantum efficiencies for the
%three color channels

pre_multipliers = compute_premultipliers(handles);

%Process all the selected images
for i = 1:length(handles.filename)
    %For plotting
    handles.xv = 0.94;
    handles.xn = 0.94;
    %"Clean" the plots from potential previous images
    axes(handles.axes_image);
    imshow(handles.xv);
    axes(handles.axes_nir);
    imshow(handles.xn);
    
    
    %LOAD THE IMAGE
    updateInfo(handles, ['Processing image ' num2str(i) ' of ' num2str(length(handles.filename)) '... ']);
    
    switch handles.input_type
        case 'vis'
            handles.xv = load_image(handles.filename{i}, handles.pathname, 'vis', cfa_ind, handles.nir_info, handles.cfa_pattern, pre_multipliers(1,:));
        case 'nir'
            handles.xn = load_image(handles.filename{i}, handles.pathname, 'nir', cfa_ind, handles.nir_info, handles.cfa_pattern, pre_multipliers(1,:));  
        case 'visnir'
            handles.xv = load_image(handles.filename{i}, handles.pathname, 'visnir_vis', cfa_ind, handles.nir_info, handles.cfa_pattern, pre_multipliers(1,:));
            handles.xn = load_image(handles.filename{i}, handles.pathname, 'visnir_nir', cfa_ind, handles.nir_info, handles.cfa_pattern, pre_multipliers(2,:));
        case 'tworaw'
            handles.xv = load_image(handles.filename{i}, handles.pathname, 'vis', cfa_ind, handles.nir_info, handles.cfa_pattern, pre_multipliers(1,:));
            handles.xn = load_image(handles.filename{i}, handles.pathname2, 'nir', cfa_ind, handles.nir_info, handles.cfa_pattern, pre_multipliers(2,:));
    end

    updateInfo(handles, ['Image ' handles.filename{i} ' successfully loaded...']);  

    %If the camera rig of the lab is used, we need to rotate the nir image
    %by 180 Degrees
    if(handles.rotate_image == 1)
        %Rotate the nir image by 180 degrees (2 times 90)
        handles.xn = rot90(handles.xn, 2);
    end
    
    
    %IMAGE ALIGNMENT (ONLY ON MAC!)
    if(handles.align_images == 1)
        updateInfo(handles, 'Aligning images...')
        %Subsample the mosaiced image by a factor of 2 in both dimensions
        [r, c] = find(handles.xv(1:2, 1:2, getIndexFromString('R')) ~= 0, 1);
        [r_nir, c_nir] = find(handles.xn(1:2, 1:2, getIndexFromString(handles.nir_info)) ~= 0, 1);
        
        imwrite(handles.xv(r:2:end, c:2:end, getIndexFromString('R'))./max(max(handles.xv(r:2:end, c:2:end, getIndexFromString('R')))), './imalign/im1_red.jpg', 'jpg', 'Quality', 100);
        imwrite(handles.xn(r_nir:2:end, c_nir:2:end, getIndexFromString(handles.nir_info))./max(max(handles.xn(r_nir:2:end, c_nir:2:end, getIndexFromString(handles.nir_info)))), './imalign/im2.jpg', 'jpg', 'Quality', 100);
        
        %Use Matthew Brown's image alignment code (works only on Mac)
        unix('./imalign/imalign im1 ./imalign/im1_red.jpg im2 ./imalign/im2.jpg');
        %Now the nir channel has been aligned to the red channel of the visible
        %spectrum, and the cropping rectangle (which tells us which information
        %has been cropped in the red channel) has been stored. We recover the
        %coordinates and crop the mosaiced visible information. We also load
        %the aligned nir channel.
        
        %Read the coordinates
        coords = textread('./imalign/rect_coords.txt');
        %Crop the visible channels
        handles.xv = handles.xv(2*coords(1)+1:2*(coords(3)+1), 2*coords(2)+1:2*(coords(4)+1), :);
        
        %Load the aligned nir image
        toto = imread('./imalign/nir.tif');
        [m,n,~] = size(toto);
        
        handles.xn = zeros(2*m, 2*n);
        handles.xn(r:2:end, r:2:end) = toto(:,:,1);
        
        updateInfo(handles, 'Image has been aligned...')
    end
       
    if(handles.show_images == 1)
        show_images(handles, handles.xv./max(max(handles.xv(:))), handles.xn./max(max(handles.xn(:))));
    end
    
    updateInfo(handles, ['Processing image ' handles.filename{i} '...']);
    
    
    %DEMOSAICING AND WHITE BALANCING
    if(strcmp(handles.order, 'D-W'))
        % Demosaicing -> White Balancing
        %Demosaicing step
        updateInfo(handles, ['Demosaicing the image using ' handles.demosaicing_method '...']);
        switch handles.input_type
            case 'vis'
                handles.xv = apply_demosaicing(handles, 'vis');
                %White balancing step
                updateInfo(handles, ['White balancing the image using ' handles.wbMode ' algorithm...']);
                handles.xv = apply_white_balancing(handles.xv, handles.xn, handles.wbMode);
                handles.xv = handles.xv./max(max(handles.xv(:)));
            case 'nir'
                handles.xn = apply_demosaicing(handles, 'nir_only');
                handles.xn = handles.xn./max(max(handles.xn(:)));
            case {'visnir', 'tworaw'}
                handles.xv = apply_demosaicing(handles, 'vis');
                handles.xn = apply_demosaicing(handles, 'nir');
                
                %White balancing step
                updateInfo(handles, ['White balancing the image using ' handles.wbMode ' algorithm... ']);
                handles.xv = apply_white_balancing(handles.xv, handles.xn, handles.wbMode);
                
                handles.xv = handles.xv./max(max(handles.xv(:)));
                handles.xn = handles.xn./max(max(handles.xn(:)));
        end

    elseif(strcmp(handles.order, 'W-D'))
        % White Balancing -> Demosaicing
        switch handles.input_type
            case 'vis'
                %White balancing step
                updateInfo(handles, ['White balancing the image using ' handles.wbMode ' algorithm... ']);
                
                handles.xv = apply_white_balancing(handles.xv, handles.xn, handles.wbMode, handles.cfa_pattern);               
                %Demosaicing step
                updateInfo(handles, ['Demosaicing the image using ' handles.demosaicing_method '...']);
                handles.xv = apply_demosaicing(handles, 'vis');
                handles.xv = handles.xv./max(max(handles.xv(:)));
            case 'nir'
                %Demosaicing step
                updateInfo(handles, ['Demosaicing the image using ' handles.demosaicing_method '...']);
                handles.xn = apply_demosaicing(handles, 'nir_only');
                handles.xn = handles.xn./max(max(handles.xn(:)));                                                
            case {'visnir', 'tworaw'}
                %White balancing step
                updateInfo(handles, ['White balancing the image using ' handles.wbMode ' algorithm... ']);
                handles.xv = apply_white_balancing(handles.xv, handles.xn, handles.wbMode, handles.cfa_pattern);
                
                %Demosaicing step
                updateInfo(handles, ['Demosaicing the image using ' handles.demosaicing_method '...']);
                handles.xv = apply_demosaicing(handles, 'vis');
                handles.xn = apply_demosaicing(handles, 'nir');
                
                handles.xv = handles.xv./max(max(handles.xv(:)));
                handles.xn = handles.xn./max(max(handles.xn(:)));
        end
    else
        error('The order must be either D-W or W-D...')
    end
    
    if(handles.show_images == 1)
        show_images(handles, handles.xv, handles.xn);
    end
       
    %%%POSSIBLE TODO: DENOISING
      
    
    %CONVERSION TO SRGB

    if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
        updateInfo(handles, 'Converting image to sRGB color space... ');
        %Convert from camera RGB (cRGB) to sRGB
        handles.xv = cRGB2sRGB(handles.xv , handles.cRGB2sRGBmat);
        if(handles.show_images == 1)
            show_images(handles, handles.xv, handles.xn);
        end
    end
    
    
    %TONE MAPPING
    updateInfo(handles, ['Applying ' lower(handles.tonemapping_method) ' tonemapping...']);
    
    switch handles.input_type
        case 'vis'
            handles.xv = apply_tonemapping(handles.xv, handles.tonemapping_method, 'vis');
        case 'nir'
            handles.xn = apply_tonemapping(handles.xn, handles.tonemapping_method, 'nir');
        case {'visnir', 'tworaw'}
            handles.xv = apply_tonemapping(handles.xv, handles.tonemapping_method, 'vis');
            handles.xn = apply_tonemapping(handles.xn, handles.tonemapping_method, 'nir');
    end
    if(handles.show_images == 1)
        show_images(handles, handles.xv, handles.xn);
    end
    
    %Update handles structure
    guidata(hObject, handles);
    
    if(~strcmp(handles.tonemapping_method, 'Colorimetric reproduction'))
        
        %BLACK-/WHITEPOINT CORRECTION AND GLOBAL GAMMA CORRECTION
        updateInfo(handles, 'Performing B/W point correction and global gamma correction');
        if(strcmp(handles.gamma, 'variable'))
            %Get the actual values from the GUI
            gamma_value = get(handles.slider_gamma, 'Value');
            lower_quantile = str2double(get(handles.edit_quantileLower, 'String'));
            upper_quantile = str2double(get(handles.edit_quantileUpper, 'String'));
            
            updateInfo(handles, 'Please select gamma value and lower/upper quantile...');
            if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
                preview = apply_bw_correction(imresize(handles.xv, [400 NaN]), lower_quantile, upper_quantile, gamma_value);
                axes(handles.axes_image);
                imshow(preview)
            end
            if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
                preview = apply_bw_correction(imresize(handles.xn, [400 NaN]), lower_quantile, upper_quantile, gamma_value);
                axes(handles.axes_nir);
                imshow(preview)
            end
            set(handles.pushbutton_setGamma, 'Visible', 'on');
            
            %Wait until the "Set" button is pressed, upon which uiresume is
            %called
            uiwait
        end
        
        %Update handles structure
        guidata(hObject, handles);
        
        %Get the actual values from the GUI
        gamma_value = get(handles.slider_gamma, 'Value');
        lower_quantile = str2double(get(handles.edit_quantileLower, 'String'));
        upper_quantile = str2double(get(handles.edit_quantileUpper, 'String'));
        
        if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
            handles.xv = apply_bw_correction(handles.xv, lower_quantile, upper_quantile, gamma_value);
        end
        if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
            handles.xn = apply_bw_correction(handles.xn, lower_quantile, upper_quantile, gamma_value);
        end
        
        if(handles.show_images == 1)
            show_images(handles, handles.xv, handles.xn);
        end
    end
    
    %SAVING
    updateInfo(handles, 'Saving image... ');
    
    path = get(handles.edit_OutputDir, 'String');
    %Make sure the path ends with a '/'
    if(~strcmp(path(end), '/'))
        path = [path '/'];
    end
    
    %Get the filename without file extension
    foo = handles.filename{i};
    filenameBase = foo(1:strfind(foo, '.')-1);
    
    %Check if the directory where the images are to be stored already exists
    %and create it if necessary
    if(~exist(path, 'dir'))
        updateInfo(handles, 'Directory does not exist. Creating it for you...');
        mkdir(path)
    end
    
    switch handles.output_format
        case 'png'
            if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
                imwrite(handles.xv, [path filenameBase '_vis.png']);
            end
            if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
                imwrite(handles.xn./max(max(handles.xn(:))), [path filenameBase '_nir.png']);
            end
        case 'tiff'
            if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
                imwrite(handles.xv, [path filenameBase '_vis.tiff']);
            end
            if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
                imwrite(handles.xn./max(max(handles.xn(:))), [path filenameBase '_nir.tiff']);
            end
        case 'jpg'
            if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
                imwrite(handles.xv, [path filenameBase '_vis.jpg'], 'Quality', 95);
            end
            if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
                imwrite(handles.xn./max(max(handles.xn(:))), [path filenameBase '_nir.jpg'], 'Quality', 95);
            end
        otherwise
            error('Output format does not exist...')
    end
    
    updateInfo(handles, ['Image ' handles.filename{i} ' saved...']);
end

updateInfo(handles, 'All images have been processed.')
set(handles.button_startProcessing, 'Enable', 'on');

%Update handles structure
guidata(hObject, handles);


%Function used to update the infobox text
function updateInfo(handles, new_info_line)
info_text = cellstr(get(handles.edit_info, 'String'));
[counter, ~] = size(info_text);
for i = counter+1:-1:2
    info_text{i} = info_text{i-1};
end
info_text{1} = char(new_info_line);
set(handles.edit_info,'String',info_text);
drawnow

function show_images(handles, xv, xn)
%Display the images
if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
    axes(handles.axes_image);
    imshow(xv);
end
if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
    axes(handles.axes_nir);
    imshow(xn);
end
drawnow

% --- Executes on selection change in popupmenu_cameraSelectionVIS.
function popupmenu_cameraSelectionVIS_Callback(hObject, eventdata, handles)
contents = cellstr(get(hObject, 'String'));
handles.cameramodelCam1 =  contents{get(hObject, 'Value')};

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in pushbutton_addCamera.
function pushbutton_addCamera_Callback(hObject, eventdata, handles)
%Open the camera GUI
addNewCameraGUI


% --- Executes on button press in radiobutton_red.
function radiobutton_red_Callback(hObject, eventdata, handles)
set(handles.radiobutton_red, 'Value', 1);
set(handles.radiobutton_green, 'Value', 0);
set(handles.radiobutton_blue, 'Value', 0);
set(handles.radiobutton_all, 'Value', 0);
handles.nir_info = 'R';

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_green.
function radiobutton_green_Callback(hObject, eventdata, handles)
set(handles.radiobutton_green, 'Value', 1);
set(handles.radiobutton_red, 'Value', 0);
set(handles.radiobutton_blue, 'Value', 0);
set(handles.radiobutton_all, 'Value', 0);
handles.nir_info = 'G';

% Update handles structure
guidata(hObject, handles);

% --- Executes on button press in radiobutton_blue.
function radiobutton_blue_Callback(hObject, eventdata, handles)
set(handles.radiobutton_blue, 'Value', 1);
set(handles.radiobutton_red, 'Value', 0);
set(handles.radiobutton_green, 'Value', 0);
set(handles.radiobutton_all, 'Value', 0);
handles.nir_info = 'B';

% Update handles structure
guidata(hObject, handles);

% --- Executes on button press in radiobutton_all.
function radiobutton_all_Callback(hObject, eventdata, handles)
set(handles.radiobutton_blue, 'Value', 0);
set(handles.radiobutton_red, 'Value', 0);
set(handles.radiobutton_green, 'Value', 0);
set(handles.radiobutton_all, 'Value', 1);
handles.nir_info = 'all';

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_singleRAW.
function radiobutton_singleRAW_Callback(hObject, eventdata, handles)
set(handles.radiobutton_singleRAW, 'Value', 1);
set(handles.radiobutton_twoRAW, 'Value', 0);
if(strcmp(handles.singletype, 'vis'))
    set(handles.uipanel_nirInfo, 'Visible', 'off');
end
set(handles.uipanel_alignImages, 'Visible', 'off');
set(handles.checkbox_alignImages, 'Value', 0);
set(handles.text23, 'Visible', 'off');

set(handles.radiobutton_VIS, 'Value', 1);
set(handles.radiobutton_NIR, 'Value', 0);
set(handles.radiobutton_VISNIR, 'Value', 0);
set(handles.text21, 'String', 'VIS');
handles.input_type = 'vis';

set(handles.popupmenu_cameraSelectionNIR, 'Visible', 'off');

set(handles.radiobutton_VIS, 'Visible', 'on');
set(handles.radiobutton_NIR, 'Visible', 'on');
set(handles.radiobutton_VISNIR, 'Visible', 'on');
set(handles.text_cfatype, 'Visible', 'on');

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_twoRAW.
function radiobutton_twoRAW_Callback(hObject, eventdata, handles)
set(handles.radiobutton_twoRAW, 'Value', 1);
set(handles.radiobutton_singleRAW, 'Value', 0);
handles.input_type = 'tworaw';

set(handles.uipanel_nirInfo, 'Visible', 'on');
set(handles.uipanel_alignImages, 'Visible', 'on');
set(handles.text21, 'String', 'VIS');
set(handles.text23, 'Visible', 'on');
set(handles.popupmenu_cameraSelectionNIR, 'Visible', 'on');

set(handles.radiobutton_VIS, 'Visible', 'off');
set(handles.radiobutton_NIR, 'Visible', 'off');
set(handles.radiobutton_VISNIR, 'Visible', 'off');
set(handles.text_cfatype, 'Visible', 'off');

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_demosaicfirst.
function radiobutton_demosaicfirst_Callback(hObject, eventdata, handles)
set(handles.radiobutton_demosaicfirst, 'Value', 1);
set(handles.radiobutton_WBfirst, 'Value', 0);
handles.order = 'D-W';

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_WBfirst.
function radiobutton_WBfirst_Callback(hObject, eventdata, handles)
set(handles.radiobutton_WBfirst, 'Value', 1);
set(handles.radiobutton_demosaicfirst, 'Value', 0);
handles.order = 'W-D';

% Update handles structure
guidata(hObject, handles);


% --- Executes on selection change in popupmenu_wbmode.
function popupmenu_wbmode_Callback(hObject, eventdata, handles)
contents = cellstr(get(hObject, 'String'));
handles.wbMode =  contents{get(hObject, 'Value')};

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_spectrumVISNIR.
function radiobutton_spectrumVISNIR_Callback(hObject, eventdata, handles)
set(handles.radiobutton_spectrumVISNIR, 'Value', 1);
set(handles.radiobutton_spectrumNIR, 'Value', 0);
set(handles.radiobutton_spectrumVIS, 'Value', 0);
handles.spectrum = 'vis_nir';

% Update handles structure
guidata(hObject, handles);

% --- Executes on button press in radiobutton_spectrumNIR.
function radiobutton_spectrumNIR_Callback(hObject, eventdata, handles)
set(handles.radiobutton_spectrumNIR, 'Value', 1);
set(handles.radiobutton_spectrumVISNIR, 'Value', 0);
set(handles.radiobutton_spectrumVIS, 'Value', 0);
handles.spectrum = 'nir';

% Update handles structure
guidata(hObject, handles);

% --- Executes on button press in radiobutton_spectrumVIS.
function radiobutton_spectrumVIS_Callback(hObject, eventdata, handles)
set(handles.radiobutton_spectrumVIS, 'Value', 1);
set(handles.radiobutton_spectrumVISNIR, 'Value', 0);
set(handles.radiobutton_spectrumNIR, 'Value', 0);
handles.spectrum = 'vis';

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in pushbutton_updateCameraList.
function pushbutton_updateCameraList_Callback(hObject, eventdata, handles)
load cameras.mat
%Store the cell containing the camera models along with their 3x3 cRGB2XYZ
%transformation matrices
handles.cameras = mm;

set(handles.popupmenu_cameraSelectionVIS, 'String', handles.cameras(:,1));
set(handles.popupmenu_cameraSelectionNIR, 'String', handles.cameras(:,1));

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in pushbutton_rename.
function pushbutton_rename_Callback(hObject, eventdata, handles)
Rename_images


% --- Executes on button press in checkbox_alignImages.
function checkbox_alignImages_Callback(hObject, eventdata, handles)
state = get(hObject,'Value');
if(state == 1)
    handles.align_images = 1;
else
    handles.align_images = 0;
end

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in checkbox_rotateImage.
function checkbox_rotateImage_Callback(hObject, eventdata, handles)
state = get(hObject,'Value');
if(state == 1)
    handles.rotate_image = 1;
else
    handles.rotate_image = 0;
end

% Update handles structure
guidata(hObject, handles);


% --- Executes on selection change in popupmenu_demosaicingMethod.
function popupmenu_demosaicingMethod_Callback(hObject, eventdata, handles)
contents = cellstr(get(hObject, 'String'));
handles.demosaicing_method =  contents{get(hObject, 'Value')};

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_PNG.
function radiobutton_PNG_Callback(hObject, eventdata, handles)
set(handles.radiobutton_PNG, 'Value', 1);
set(handles.radiobutton_TIFF, 'Value', 0);
set(handles.radiobutton_JPEG, 'Value', 0);
handles.output_format = 'png';

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_JPEG.
function radiobutton_JPEG_Callback(hObject, eventdata, handles)
set(handles.radiobutton_PNG, 'Value', 0);
set(handles.radiobutton_TIFF, 'Value', 0);
set(handles.radiobutton_JPEG, 'Value', 1);
handles.output_format = 'jpg';

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_TIFF.
function radiobutton_TIFF_Callback(hObject, eventdata, handles)
set(handles.radiobutton_PNG, 'Value', 0);
set(handles.radiobutton_TIFF, 'Value', 1);
set(handles.radiobutton_JPEG, 'Value', 0);
handles.output_format = 'tiff';

% Update handles structure
guidata(hObject, handles);

% --- Executes on button press in radiobutton_VIS.
function radiobutton_VIS_Callback(hObject, eventdata, handles)
set(handles.radiobutton_VIS, 'Value', 1);
set(handles.radiobutton_NIR, 'Value', 0);
set(handles.radiobutton_VISNIR, 'Value', 0);

handles.input_type = 'vis';
set(handles.text21, 'String', 'VIS');

set(handles.uipanel_nirInfo, 'Visible', 'off');

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_NIR.
function radiobutton_NIR_Callback(hObject, eventdata, handles)
set(handles.radiobutton_VIS, 'Value', 0);
set(handles.radiobutton_NIR, 'Value', 1);
set(handles.radiobutton_VISNIR, 'Value', 0);

handles.input_type = 'nir';
set(handles.text21, 'String', 'NIR');
set(handles.uipanel_nirInfo, 'Visible', 'on');

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in radiobutton_VISNIR.
function radiobutton_VISNIR_Callback(hObject, eventdata, handles)
set(handles.radiobutton_VIS, 'Value', 0);
set(handles.radiobutton_NIR, 'Value', 0);
set(handles.radiobutton_VISNIR, 'Value', 1);

handles.input_type = 'visnir';
set(handles.text21, 'String', 'VIS+NIR');

set(handles.uipanel_nirInfo, 'Visible', 'on');

% Update handles structure
guidata(hObject, handles);

% --- Executes on selection change in popupmenu_tonemapping.
function popupmenu_tonemapping_Callback(hObject, eventdata, handles)
contents = cellstr(get(hObject, 'String'));
handles.tonemapping_method =  contents{get(hObject, 'Value')};

% Update handles structure
guidata(hObject, handles);


% --- Executes on selection change in popupmenu_cameraSelectionNIR.
function popupmenu_cameraSelectionNIR_Callback(hObject, eventdata, handles)
contents = cellstr(get(hObject, 'String'));
handles.cameramodelCam2 =  contents{get(hObject, 'Value')};

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in checkbox_showResults.
function checkbox_showResults_Callback(hObject, eventdata, handles)
state = get(hObject,'Value');
if(state == 1)
    handles.show_images = 1;
else
    handles.show_images = 0;
end

% Update handles structure
guidata(hObject, handles);

% --- Executes on slider movement.
function slider_gamma_Callback(hObject, eventdata, handles)
handles.gamma_value = get(hObject, 'Value');
set(handles.edit_gammavalue, 'String', num2str(handles.gamma_value));

if(strcmp(handles.gamma, 'variable'))
    if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
        preview = apply_bw_correction(imresize(handles.xv, [400 NaN]), handles.lower_quantile, handles.upper_quantile, handles.gamma_value);
        axes(handles.axes_image);
        imshow(preview)
    end
    if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
        preview = apply_bw_correction(imresize(handles.xn, [400 NaN]), handles.lower_quantile, handles.upper_quantile, handles.gamma_value);
        axes(handles.axes_nir);
        imshow(preview)
    end
    drawnow
end

% Update handles structure
guidata(hObject, handles);

function edit_gammavalue_Callback(hObject, eventdata, handles)
handles.gamma_value = str2double(get(hObject,'String'));
set(handles.slider_gamma, 'Value', handles.gamma_value);

if(strcmp(handles.gamma, 'variable'))
    if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
        preview = apply_bw_correction(imresize(handles.xv, [400 NaN]), handles.lower_quantile, handles.upper_quantile, handles.gamma_value);
        axes(handles.axes_image);
        imshow(preview)
    end
    if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
        preview = apply_bw_correction(imresize(handles.xn, [400 NaN]), handles.lower_quantile, handles.upper_quantile, handles.gamma_value);
        axes(handles.axes_nir);
        imshow(preview)
    end
    drawnow
end

% Update handles structure
guidata(hObject, handles);

% --- Executes on button press in radiobutton_gammaVariable.
function radiobutton_gammaVariable_Callback(hObject, eventdata, handles)
set(handles.radiobutton_gammaVariable, 'Value', 1);
set(handles.radiobutton_gammaFixed, 'Value', 0);
handles.gamma = 'variable';
handles.continue_workflow = 0;

set(handles.pushbutton_setGamma, 'Visible', 'off');

% Update handles structure
guidata(hObject, handles);

% --- Executes on button press in radiobutton_gammaFixed.
function radiobutton_gammaFixed_Callback(hObject, eventdata, handles)
set(handles.radiobutton_gammaVariable, 'Value', 0);
set(handles.radiobutton_gammaFixed, 'Value', 1);
handles.gamma = 'fixed';
handles.continue_workflow = 1;

set(handles.pushbutton_setGamma, 'Visible', 'off');

% Update handles structure
guidata(hObject, handles);


% --- Executes on button press in pushbutton_setGamma.
function pushbutton_setGamma_Callback(hObject, eventdata, handles)
uiresume
set(handles.pushbutton_setGamma, 'Visible', 'off');

% Update handles structure
guidata(hObject, handles);



function edit_OutputDir_Callback(hObject, eventdata, handles)

function info_Callback(hObject, eventdata, handles)

function edit_info_Callback(hObject, eventdata, handles)

function edit_quantileLower_Callback(hObject, eventdata, handles)
new_quant = str2num(get(hObject,'String'));

%Make sure the value entered is valid, which is the case if str2num does
%not return an empty cell.
if(~isempty(new_quant))
    set(hObject,'String', get(hObject,'String'));
    handles.lower_quantile = new_quant;
else
    set(hObject,'String', handles.lower_quantile);
end


if(strcmp(handles.gamma, 'variable'))
    if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
        preview = apply_bw_correction(imresize(handles.xv, [400 NaN]), handles.lower_quantile, handles.upper_quantile, handles.gamma_value);
        axes(handles.axes_image);
        imshow(preview)
    end
    if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
        preview = apply_bw_correction(imresize(handles.xn, [400 NaN]), handles.lower_quantile, handles.upper_quantile, handles.gamma_value);
        axes(handles.axes_nir);
        imshow(preview)
    end
    drawnow
end

% Update handles structure
guidata(hObject, handles);


function edit_quantileUpper_Callback(hObject, eventdata, handles)
new_quant = str2num(get(hObject,'String'));

%Make sure the value entered is valid, which is the case if str2num does
%not return an empty cell.
if(~isempty(new_quant))
    set(hObject,'String', get(hObject,'String'));
    handles.upper_quantile = new_quant;
else
    set(hObject,'String', num2str(handles.upper_quantile));
end

if(strcmp(handles.gamma, 'variable'))
    if(strcmp(handles.input_type, 'vis') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
        preview = apply_bw_correction(imresize(handles.xv, [400 NaN]), handles.lower_quantile, handles.upper_quantile, handles.gamma_value);
        axes(handles.axes_image);
        imshow(preview)
    end
    if(strcmp(handles.input_type, 'nir') || strcmp(handles.input_type, 'visnir') || strcmp(handles.input_type, 'tworaw'))
        preview = apply_bw_correction(imresize(handles.xn, [400 NaN]), handles.lower_quantile, handles.upper_quantile, handles.gamma_value);
        axes(handles.axes_nir);
        imshow(preview)
    end
    drawnow
end

% Update handles structure
guidata(hObject, handles);


% --- Outputs from this function are returned to the command line.
function varargout = workflow_gui_OutputFcn(hObject, eventdata, handles)
varargout{1} = handles.output;

%--------------------------------------------------------------------------------------------
%Create functions
%--------------------------------------------------------------------------------------------
% --- Executes during object creation, after setting all properties.
function popupmenu_cameraSelectionNIR_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function popupmenu_tonemapping_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function edit_filenameBase_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function popupmenu_demosaicingMethod_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function popupmenu_wbmode_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function edit_info_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function info_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

function edit_OutputDir_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function popupmenu_cameraSelectionVIS_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function edit_quantileLower_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function edit_quantileUpper_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end

% --- Executes during object creation, after setting all properties.
function slider_gamma_CreateFcn(hObject, eventdata, handles)
if isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor',[.9 .9 .9]);
end

% --- Executes during object creation, after setting all properties.
function edit_gammavalue_CreateFcn(hObject, eventdata, handles)
if ispc && isequal(get(hObject,'BackgroundColor'), get(0,'defaultUicontrolBackgroundColor'))
    set(hObject,'BackgroundColor','white');
end
